import logging

import numpy as np

from autolens import exc
from autolens.imaging import imaging_util
from autolens.imaging import scaled_array

logging.basicConfig()
logger = logging.getLogger(__name__)


class Mask(scaled_array.ScaledSquarePixelArray):
    """
    A mask represented by an ndarray where True is masked.
    """

    def __getitem__(self, coords):
        try:
            return super(Mask, self).__getitem__(coords)
        except IndexError:
            return True

    @classmethod
    def padded_for_shape_and_pixel_scale(cls, shape, pixel_scale):
        """
        Setup the mask such that all pixels are padded.

        Parameters
        ----------
        shape : (int, int)
            The (x,y) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(np.full(tuple(map(lambda d: int(d), shape)), False, dtype='bool'), pixel_scale)

    @classmethod
    def masked_for_shape_and_pixel_scale(cls, shape, pixel_scale):
        """
        Setup the mask such that all pixels are masked.

        Parameters
        ----------
        shape : (int, int)
            The (x,y) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(np.full(tuple(map(lambda d: int(d), shape)), True, dtype='bool'), pixel_scale)

    @classmethod
    def circular(cls, shape, pixel_scale, radius_mask_arcsec, centre=(0., 0.)):
        """
        Setup the mask as a circle, using a specified arc second radius.

        Parameters
        ----------
        shape: (int, int)
            The (x,y) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        radius_mask_arcsec : float
            The radius of the circular mask in arc seconds.
        centre: (float, float)
            The centre of the mask.
        """
        mask = imaging_util.mask_circular_from_shape_pixel_scale_and_radius(shape, pixel_scale, radius_mask_arcsec,
                                                                            centre)
        return cls(mask.astype('bool'), pixel_scale)

    @classmethod
    def annular(cls, shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec, centre=(0., 0.)):
        """
        Setup the mask as an annulus, using a specified inner and outer radius in arc seconds.

        Parameters
        ----------
        shape : (int, int)
            The (x,y) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        outer_radius_arcsec : float
            The inner radius of the annulus mask in arc seconds.
        outer_radius_arcsec : float
            The outer radius of the annulus mask in arc seconds.
        centre: (float, float)
            The centre of the mask.
        """
        mask = imaging_util.mask_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec,
                                                                          outer_radius_arcsec, centre)
        return cls(mask.astype('bool'), pixel_scale)

    @classmethod
    def anti_annular(cls, shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec, outer_radius_2_arcsec,
                     centre=(0., 0.)):
        """
        Setup the mask as an annulus, using a specified inner and outer radius in arc seconds.

        Parameters
        ----------
        shape : (int, int)
            The (x,y) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The inner radius of the annulus mask in arc seconds.
        outer_radius_arcsec : float
            The outer radius of the annulus mask in arc seconds.
        outer_radius_2_arcsec
        centre: (float, float)
            The centre of the mask.
        """
        mask = imaging_util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec,
                                                                               outer_radius_arcsec,
                                                                               outer_radius_2_arcsec, centre)
        return cls(mask, pixel_scale)

    @property
    def pixels_in_mask(self):
        return int(np.size(self) - np.sum(self))

    @property
    def grid_to_pixel(self):
        """A 1D array of mappings between every padded pixel and its 2D pixel coordinates."""
        return imaging_util.grid_to_pixel_from_mask(self).astype('int')

    def map_2d_array_to_masked_1d_array(self, array_2d):
        """For a 2D data-array (e.g. the _data, noise-mappers, etc.) mappers it to a masked 1D array of values usin this mask.

        Parameters
        ----------
        array_2d : ndarray | None | float
            The data to be mapped to a masked 1D array.
        """
        if array_2d is None or isinstance(array_2d, float):
            return array_2d
        return imaging_util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(self, array_2d)

    @imaging_util.Memoizer()
    def blurring_mask_for_psf_shape(self, psf_shape):
        """Compute the blurring mask, which represents all masked pixels whose light will be blurred into padded \
        pixels via PSF convolution.

        Parameters
        ----------
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region (e.g. the shape of the PSF)
        """

        if psf_shape[0] % 2 == 0 or psf_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = imaging_util.mask_blurring_from_mask_and_psf_shape(self, psf_shape)

        return Mask(blurring_mask, self.pixel_scale)

    @property
    def border_pixels(self):
        """The indicies of the mask's border pixels, where a border pixel is a pixel inside the mask but on its edge \
        (next to at least one pixel with a *True* value).
        """
        return imaging_util.border_pixels_from_mask(self).astype('int')

    def border_sub_pixels(self, sub_grid_size):
        """The indicies of the mask's border sub-pixels, where a border sub-pixel is the sub-pixel in a mask border \
        _data-pixel which is closest to the edge."""
        return imaging_util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=self,
                                                                                       pixel_scales=self.pixel_scales,
                                                                                       sub_grid_size=sub_grid_size).astype(
            'int')


class ImagingGrids(object):

    def __init__(self, image, sub, blurring):
        """The grids containing the (x,y) arc-second coordinates of padded pixels in a mask. There are 3 grids:

        The _data-grid - the (x,y) coordinate at the center of every padded pixel.
        The sub-grid - the (x,y) coordinates of every sub-pixel in every padded pixel, each using a grid of size \
        sub_grid_size x sub_grid_size.
        The blurring-grid - the (x,y) coordinates of all blurring pixels, which are masked pixels whose light is \
        blurred into masked pixels during PSF convolution.

        The grids are stored as 1D arrays, where each entry corresponds to an padded (sub-)pixel. The 1D array is \
        ordered such pixels begin from the top-row of the masks 2D array and then downwards.

        Parameters
        -----------
        image : ImageGrid
            The grid of (x,y) arc-second coordinates at the centre of every padded pixel.
        sub : SubGrid
            The grid of (x,y) arc-second coordinates at the centre of every padded pixel's sub-pixels.
        blurring : ImageGrid | ndarray | None
            The grid of (x,y) arc-second coordinates at the centre of every blurring-mask pixel.
        """
        self.image = image
        self.sub = sub
        self.blurring = blurring

    @classmethod
    def grids_from_mask_sub_grid_size_and_psf_shape(cls, mask, sub_grid_size, psf_shape):
        """Setup the *ImagingGrids* from a mask, sub-grid size and psf-shape.

        Parameters
        -----------
        mask : Mask
            The mask whose padded pixels the imaging-grids are setup using.
        sub_grid_size : int
            The size of a sub-pixels sub-grid (sub_grid_size x sub_grid_size).
        psf_shape : (int, int)
            the shape of the PSF used in the analysis, which defines the mask's blurring-region.
        """
        image_grid = ImageGrid.from_mask(mask)
        sub_grid = SubGrid.from_mask_and_sub_grid_size(mask, sub_grid_size)
        blurring_grid = ImageGrid.blurring_grid_from_mask_and_psf_shape(mask, psf_shape)
        return ImagingGrids(image_grid, sub_grid, blurring_grid)

    @classmethod
    def from_shape_and_pixel_scale(cls, shape, pixel_scale, sub_grid_size=2, psf_shape=(1, 1)):
        image_grid = ImageGrid.from_shape_and_pixel_scale(shape=shape, pixel_scale=pixel_scale)
        sub_grid = SubGrid.from_shape_pixel_scale_and_sub_grid_size(shape=shape, pixel_scale=pixel_scale,
                                                                    sub_grid_size=sub_grid_size)
        blurring_grid = np.array([[0.0, 0.0]])
        return ImagingGrids(image_grid, sub_grid, blurring_grid)

    @classmethod
    def padded_grids_from_mask_sub_grid_size_and_psf_shape(cls, mask, sub_grid_size, psf_shape):
        """Setup the collection of padded imaging-grids from a mask, using also an input sub-grid size to setup the \
        sub-grid and psf-shape to setup the blurring grid.

        Parameters
        -----------
        mask : Mask
            The mask whose padded pixels the imaging-grids are setup using.
        sub_grid_size : int
            The size of a sub-pixels sub-grid (sub_grid_size x sub_grid_size).
        psf_shape : (int, int)
            the shape of the PSF used in the analysis, which therefore defines the mask's blurring-region.
        """
        image_padded_grid = PaddedImageGrid.padded_grid_from_shapes_and_pixel_scale(shape=mask.shape,
                                                                                    psf_shape=psf_shape,
                                                                                    pixel_scale=mask.pixel_scale)
        sub_padded_grid = PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                          sub_grid_size=sub_grid_size,
                                                                                          psf_shape=psf_shape)
        # TODO : The blurring grid is not used when the grid mapper is called, the 0.0 0.0 stops errors inr ayT_racing
        # TODO : implement a more explicit solution
        return ImagingGrids(image=image_padded_grid, sub=sub_padded_grid, blurring=np.array([[0.0, 0.0]]))

    @classmethod
    def grids_for_simulation(cls, shape, pixel_scale, psf_shape, sub_grid_size=1):
        """Setup a collection of imaging-grids for simulating an _data of a strong lens. 
        
        This routine uses padded-grids which ensure that the PSF blurring in the simulation routine \ 
        (*imaging.PrepatoryImage.simulate*) is not degraded due to edge effects.

        Parameters
        -----------
        pixel_scale
        shape
        sub_grid_size : int
            The size of a sub-pixels sub-grid (sub_grid_size x sub_grid_size).
        psf_shape : (int, int)
            the shape of the PSF used in the analysis, which therefore defines the mask's blurring-region.
        """
        return cls.padded_grids_from_mask_sub_grid_size_and_psf_shape(mask=Mask(array=np.full(shape, False),
                                                                                pixel_scale=pixel_scale),
                                                                      sub_grid_size=sub_grid_size,
                                                                      psf_shape=psf_shape)

    def apply_function(self, func):
        if self.blurring is not None:
            return ImagingGrids(func(self.image), func(self.sub), func(self.blurring))
        else:
            return ImagingGrids(func(self.image), func(self.sub), self.blurring)

    def map_function(self, func, *arg_lists):
        return ImagingGrids(*[func(*args) for args in zip(self, *arg_lists)])

    @property
    def sub_pixels(self):
        return self.sub.shape[0]

    def __getitem__(self, item):
        return [self.image, self.sub, self.blurring][item]


class ImageGrid(np.ndarray):
    """Abstract class for a regular grid of coordinates, where each padded pixel's (x,y) arc-second coordinates \
    are represented by the value at the centre of the pixel.

    Coordinates are defined from the top-left corner, such that pixels in the top-left corner of a mask (e.g. [0,0]) \
    have a negative x and y-value in arc seconds. The mask pixel indexes are also counted from the top-left.

    An *ImageGrid* is a ndarray of shape [image_pixels, 2], where image_pixels is the total number of padded \
    _data-pixels. The first element maps of the ndarray corresponds to the padded pixel index and second element the \
    x or y arc second coordinates. For howtolens:

    - image_grid[3,1] = the 4th padded pixel's y-coordinate.
    - image_grid[6,0] = the 5th padded pixel's x-coordinate.

    Below is a visual illustration of an _data-grid, where a total of 10 pixels are padded and are included in the \
    grid.

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an howtolens mask.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|o|o|x|x|x|x|     x = True (Pixel is masked and excluded from lensing)
    |x|x|x|o|o|o|o|x|x|x|     o = False (Pixel is not masked and included in lensing)
    |x|x|x|o|o|o|o|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    The mask pixel index's will come out like this (and the direction of arc-second coordinates is highlighted
    around the mask.

    pixel_scale = 1.0"

    <--- -ve  x  +ve -->

    |x|x|x|x|x|x|x|x|x|x|  ^   image_grid[0] = [-0.5, -1.5]
    |x|x|x|x|x|x|x|x|x|x|  |   image_grid[1] = [ 0.5, -1.5]
    |x|x|x|x|x|x|x|x|x|x|  |   image_grid[2] = [-1.5, -0.5]
    |x|x|x|x|0|1|x|x|x|x| -ve  image_grid[3] = [-0.5, -0.5]
    |x|x|x|2|3|4|5|x|x|x|  y   image_grid[4] = [ 0.5, -0.5]
    |x|x|x|6|7|8|9|x|x|x| +ve  image_grid[5] = [ 1.5, -0.5]
    |x|x|x|x|x|x|x|x|x|x|  |   image_grid[6] = [-1.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x|  |   image_grid[7] = [-0.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x| \/   image_grid[8] = [ 0.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x|      image_grid[9] = [ 1.5,  0.5]
    """

    def __new__(cls, arr, mask, *args, **kwargs):
        obj = arr.view(cls)
        obj.mask = mask
        return obj

    @property
    def unlensed_grid(self):
        return ImageGrid(arr=imaging_util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=self.mask,
                                                                                          pixel_scales=self.mask.pixel_scales),
                         mask=self.mask)

    @property
    def unlensed_unmasked_grid(self):
        return ImageGrid(arr=imaging_util.image_grid_1d_from_shape_and_pixel_scales(shape=self.mask.shape,
                                                                                    pixel_scales=self.mask.pixel_scales),
                         mask=self.mask)

    @classmethod
    def from_mask(cls, mask):
        """Setup an *ImageGrid* of the regular _data-grid from a mask. The center of every padded pixel gives \
        the grid's (x,y) arc-second coordinates.

        Parameters
        -----------
        mask : Mask
            The mask whose padded pixels are used to setup the sub-pixel grids."""
        array = imaging_util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=mask.pixel_scales)
        return cls(array, mask)

    @classmethod
    def from_shape_and_pixel_scale(cls, shape, pixel_scale):
        mask = Mask.padded_for_shape_and_pixel_scale(shape=shape, pixel_scale=pixel_scale)
        array = imaging_util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=mask.pixel_scales)
        return cls(array, mask)

    @classmethod
    def blurring_grid_from_mask_and_psf_shape(cls, mask, psf_shape):
        """Setup an *ImageGrid* of the blurring-grid from a mask. The center of every padded blurring-pixel gives \
        the grid's (x,y) arc-second coordinates."""
        blurring_mask = mask.blurring_mask_for_psf_shape(psf_shape)
        return ImageGrid.from_mask(blurring_mask)

    def map_to_2d(self, array_1d):
        """ Map a 1D array the same dimension as the grid to its original 2D array.

        Parameters
        -----------
        array_1d : ndarray
            The 1D array of masked data which is mapped to 2D.
        """
        return imaging_util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d,
                                                                                               self.mask.shape,
                                                                                               self.mask.grid_to_pixel)

    def scaled_array_from_array_1d(self, array_1d):
        return scaled_array.ScaledSquarePixelArray(array=self.map_to_2d(array_1d), pixel_scale=self.mask.pixel_scale)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(ImageGrid, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(ImageGrid, self).__setstate__(state[0:-1])

    @property
    def total_pixels(self):
        return self.shape[0]

    @property
    def pixel_scale(self):

        return self.mask.pixel_scale

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing an _data-grid"""
        return np.linspace(np.min(self[:, 0]), np.max(self[:, 0]), 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing an _data-grid"""
        return np.linspace(np.min(self[:, 1]), np.max(self[:, 1]), 4)


class SubGrid(ImageGrid):
    """Abstract class for a sub-grid of coordinates. On a sub-grid, each padded pixel is sub-gridded into a uniform \
     grid of (x,y) sub-coordinates..

    Coordinates are defined from the top-left corner, such that pixels in the top-left corner of an \
    mask (e.g. [0,0]) have negative x and y-values in arc seconds. The mask pixel indexes are also counted from the \
    top-left.

    Sub-pixels follow the same convention as above, where the top-left sub-pixel has the lowest x and y-values in each \
    _data-pixel. Sub-pixel indexes include all previous sub-pixels in all previous padded _data-pixels.

    A *SubGrid* is a NumPy array of shape [image_pixels*sub_grid_pixels**2, 2]. The first element of the ndarray \
    corresponds to the padded sub-pixel index, and second element the sub-pixel's (x,y) arc second coordinates. \
    For howtolens:

    - sub_grid[9, 1] - using a 2x2 sub-grid, gives the 3rd _data-pixel's 2nd sub-pixel y-coordinate.
    - sub_grid[9, 1] - using a 3x3 sub-grid, gives the 2nd _data-pixel's 1st sub-pixel y-coordinate.
    - sub_grid[27, 0] - using a 3x3 sub-grid, gives the 4th _data-pixel's 1st sub-pixel x-coordinate.

    Below is a visual illustration of a sub grid. Like the regular grid, the indexing of each sub-pixel goes from \
    the top-left corner. In contrast to the regular grid above, our illustration below restricts the mask to just
    2 pixels, to keep the illustration brief.

    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     This is an howtolens mask.Mask, where:
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|     x = True (Pixel is masked and excluded from lensing)
    |x|x|x|x|o|o|x|x|x|x|     o = False (Pixel is not masked and included in lensing)
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|
    |x|x|x|x|x|x|x|x|x|x|

    Our regular-grid looks like it did before:

    pixel_scale = 1.0"

    <--- -ve  x  +ve -->

    |x|x|x|x|x|x|x|x|x|x|  ^
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x| +ve  image_grid[0] = [-1.5,  0.5]
    |x|x|x|0|1|x|x|x|x|x|  y   image_grid[1] = [-0.5,  0.5]
    |x|x|x|x|x|x|x|x|x|x| -ve
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x|  |
    |x|x|x|x|x|x|x|x|x|x| \/
    |x|x|x|x|x|x|x|x|x|x|

    However, we now go to each mask-pixel and derive a sub-pixel grid for it. For howtolens, for pixel 0,
    if *sub_grid_size=2*, we use a 2x2 sub-grid:

    Pixel 0 - (2x2):

           image_grid[0] = [-1.66, 0.66]
    |0|1|  image_grid[1] = [-1.33, 0.66]
    |2|3|  image_grid[2] = [-1.66, 0.33]
           image_grid[3] = [-1.33, 0.33]

    Now, we'd normally sub-grid all pixels using the same *sub_grid_size*, but for this illustration lets
    pretend we used a sub_grid_size of 3x3 for pixel 1:

             image_grid[0] = [-0.75, 0.75]
             image_grid[1] = [-0.5,  0.75]
             image_grid[2] = [-0.25, 0.75]
    |0|1|2|  image_grid[3] = [-0.75,  0.5]
    |3|4|5|  image_grid[4] = [-0.5,   0.5]
    |6|7|8|  image_grid[5] = [-0.25,  0.5]
             image_grid[6] = [-0.75, 0.25]
             image_grid[7] = [-0.5,  0.25]
             image_grid[8] = [-0.25, 0.25]
    """

    # noinspection PyUnusedLocal
    def __init__(self, array, mask, sub_grid_size=1):
        # noinspection PyArgumentList
        super(SubGrid, self).__init__()
        self.mask = mask
        self.sub_grid_size = sub_grid_size
        self.sub_grid_length = int(sub_grid_size ** 2.0)
        self.sub_grid_fraction = 1.0 / self.sub_grid_length
        self.mask = mask

    @property
    def unlensed_grid(self):
        return SubGrid(imaging_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=self.mask,
            pixel_scales=self.mask.pixel_scales,
            sub_grid_size=self.sub_grid_size),
            self.mask, self.sub_grid_size)

    @property
    def unlensed_unmasked_grid(self):
        return SubGrid(imaging_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full(self.mask.shape, False), pixel_scales=self.mask.pixel_scales,
            sub_grid_size=self.sub_grid_size),
            mask=self.mask, sub_grid_size=self.sub_grid_size)

    @classmethod
    def from_mask_and_sub_grid_size(cls, mask, sub_grid_size=1):
        """Setup a *SubGrid* of the padded _data-pixels, using a mask and a specified sub-grid size. The center of \
        every padded pixel's sub-pixels give the grid's (x,y) arc-second coordinates.

        Parameters
        -----------
        mask : Mask
            The mask whose padded pixels are used to setup the sub-pixel grids.
        sub_grid_size : int
            The size (sub_grid_size x sub_grid_size) of each _data-pixels sub-grid.
        """
        sub_grid_masked = imaging_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                                   pixel_scales=mask.pixel_scales,
                                                                                                   sub_grid_size=sub_grid_size)
        return SubGrid(sub_grid_masked, mask, sub_grid_size)

    @classmethod
    def from_shape_pixel_scale_and_sub_grid_size(cls, shape, pixel_scale, sub_grid_size):
        mask = Mask.padded_for_shape_and_pixel_scale(shape=shape, pixel_scale=pixel_scale)
        sub_grid = imaging_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=mask.pixel_scales,
                                                                                            sub_grid_size=sub_grid_size)
        return SubGrid(sub_grid, mask, sub_grid_size)

    def __array_finalize__(self, obj):
        if isinstance(obj, SubGrid):
            self.sub_grid_size = obj.sub_grid_size
            self.sub_grid_length = obj.sub_grid_length
            self.sub_grid_fraction = obj.sub_grid_fraction
            self.mask = obj.mask

    def sub_data_to_image(self, sub_array):
        """For an input sub-gridded array, mappers it from the sub-grid to a 1D _data-grid by summing each set of
        each set of sub-pixels values and dividing by the total number of sub-pixels.

        Parameters
        -----------
        sub_array : ndarray
            A 1D sub-gridded array of values (e.g. the intensities, surface-densities, potential) which is mapped to
            a 1d _data-grid array.
        """
        return np.multiply(self.sub_grid_fraction, sub_array.reshape(-1, self.sub_grid_length).sum(axis=1))

    @property
    @imaging_util.Memoizer()
    def sub_to_image(self):
        """ Compute the mapping between every sub-pixel and its host padded _data-pixel.

        For howtolens:

        - sub_to_pixel[8] = 2 -  The seventh sub-pixel is within the 3rd padded _data pixel.
        - sub_to_pixel[20] = 4 -  The nineteenth sub-pixel is within the 5th padded _data pixel.
        """
        return imaging_util.sub_to_image_from_mask(self.mask, self.sub_grid_size).astype('int')


class PaddedImageGrid(ImageGrid):

    def __new__(cls, arr, mask, image_shape, *args, **kwargs):
        """An *PaddedImageGrid* stores the (x,y) arc-second coordinates of a mask's pixels in 1D, in an analogous
        fashion to an *ImageGrid*. An *PaddedImageGrid* deviate from a normal grid in that:

        - All pixels are used (as opposed to just padded pixels)
        - The mask is padded when computing the grid, such that additional pixels beyond its edge are included.

        Padded-grids allow quantities like intensities to be computed on large 2D arrays spanning the entire _data, as
        opposed to just within the masked region.
        """
        arr = arr.view(cls)
        arr.mask = mask
        arr.image_shape = image_shape
        return arr

    @classmethod
    def padded_grid_from_shapes_and_pixel_scale(self, shape, psf_shape, pixel_scale):
        """Setup an *PaddedImageGrid* of the regular _data-grid for an input _data-shape, psf-shape and pixel-scale.

        The center of every pixel is used to setup the grid's (x,y) arc-second coordinates, including padded-pixels
        which are beyond the input shape but will have light blurred into them given the psf-shape.

        Parameters
        ----------
        pixel_scale : float
            The scale of each pixel in arc seconds
        shape : (int, int)
            The (x,y) shape of the padded-grid's 2D _data in units of pixels.
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region and therefore size of padding.
        """
        padded_shape = (shape[0] + psf_shape[0] - 1, shape[1] + psf_shape[1] - 1)
        padded_image_grid = imaging_util.image_grid_1d_masked_from_mask_and_pixel_scales(
            mask=np.full(padded_shape, False), pixel_scales=(pixel_scale, pixel_scale))
        padded_mask = Mask.padded_for_shape_and_pixel_scale(shape=padded_shape, pixel_scale=pixel_scale)
        return PaddedImageGrid(arr=padded_image_grid, mask=padded_mask, image_shape=shape)

    def convolve_array_1d_with_psf(self, padded_array_1d, psf):
        """Convolve a 2d padded array of values (e.g. intensities beforoe PSF blurring) with a PSF, and then trim \
        the convolved array to its original 2D shape.

        Parameters
        -----------
        psf
        padded_array_1d: ndarray
            A 1D array of values which were computed using the *PaddedImageGrid*.
        """
        padded_array_2d = imaging_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(padded_array_1d,
                                                                                                 self.mask.shape)
        blurred_padded_array_2d = psf.convolve(padded_array_2d)
        return imaging_util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(array_2d=blurred_padded_array_2d,
                                                                                   mask=np.full(self.mask.shape,
                                                                                                False))

    def map_to_2d(self, padded_array_1d):
        """ Map a padded 1D array of values to its origianl 2D array.

        Parameters
        -----------
        padded_array_1d : ndarray
            A 1D array of values which were computed using the *PaddedImageGrid*.
        """
        padded_array_2d = self.map_to_2d_keep_padded(padded_array_1d)
        pad_size_0 = self.mask.shape[0] - self.image_shape[0]
        pad_size_1 = self.mask.shape[1] - self.image_shape[1]
        return (padded_array_2d[pad_size_0 // 2:self.mask.shape[0] - pad_size_0 // 2,
                pad_size_1 // 2:self.mask.shape[1] - pad_size_1 // 2])

    def map_to_2d_keep_padded(self, padded_array_1d):
        """ Map a padded 1D array of values to its padded 2D array.

        Parameters
        -----------
        padded_array_1d : ndarray
            A 1D array of values which were computed using the *PaddedImageGrid*.
        """
        return imaging_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(padded_array_1d, self.mask.shape)

    @property
    def padded_shape(self):
        return self.mask.shape


class PaddedSubGrid(SubGrid, PaddedImageGrid):

    def __init__(self, arr, mask, image_shape, sub_grid_size=1):
        """A *PaddedSubGrid* stores the (x,y) arc-second coordinates of a mask's sub-pixels in 1D, in an analogous
        fashion to a *SubGrid*. A *PaddedSubGrid* deviate from a normal grid in that:

        - All pixels are used (as opposed to just padded pixels)
        - The mask is padded when computing the grid, such that additional pixels beyond its edge are included.

        Padded-grids allow quantities like intensities to be computed on large 2D arrays spanning the entire _data, as
        opposed to just within the masked region.
        """
        super(PaddedSubGrid, self).__init__(arr, mask, sub_grid_size)

    @classmethod
    def padded_grid_from_mask_sub_grid_size_and_psf_shape(cls, mask, sub_grid_size, psf_shape):
        """Setup an *PaddedSubGrid* for an input mask, sub-grid size and psf-shape.

        The center of every sub-pixel is used to setup the grid's (x,y) arc-second coordinates, including \
        padded-pixels which are beyond the input shape but will have light blurred into them given the psf-shape.

        Parameters
        ----------
        mask : Mask
            The mask whose padded pixels are used to setup the sub-pixel grids.
        sub_grid_size : int
            The size (sub_grid_size x sub_grid_size) of each _data-pixels sub-grid.
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region and therefore size of padding.
        """

        padded_shape = (mask.shape[0] + psf_shape[0] - 1, mask.shape[1] + psf_shape[1] - 1)

        padded_sub_grid = imaging_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full(padded_shape, False), pixel_scales=mask.pixel_scales, sub_grid_size=sub_grid_size)

        padded_mask = Mask.padded_for_shape_and_pixel_scale(shape=padded_shape, pixel_scale=mask.pixel_scale)

        return PaddedSubGrid(arr=padded_sub_grid, mask=padded_mask, image_shape=mask.shape,
                             sub_grid_size=sub_grid_size)

    def __array_finalize__(self, obj):
        if isinstance(obj, PaddedSubGrid):
            self.sub_grid_size = obj.sub_grid_size
            self.sub_grid_length = obj.sub_grid_length
            self.sub_grid_fraction = obj.sub_grid_fraction
            self.mask = obj.mask
            self.image_shape = obj.image_shape


class ImagingGridBorders(object):

    def __init__(self, image, sub):
        """The borders containing the pixel-index's of all padded pixels that are on the mask's border (e.g. \
        they are next to a *True* value in at least one of the surrounding 8 pixels). There are 2 borders:

        The _data-border - The pixel-indexes of all padded pixels in an *ImageGrid* that are border-pixels.
        The sub-grid - The sub-pixel-indexes of all padded pixels in a *SubGrid* that are border-pixels.

        A polynomial is fitted to the (x,y) coordinates of each border's pixels. This allows us to relocate \
        demagnified pixel's in a grid to its border, so that they do not disrupt an adaptive inversion.

        Parameters
        -----------
        image : ImageGridBorder
            The pixel-index's of all _data-pixels which are on the mask border.
        sub : SubGridBorder
            The sub-pixel-index's of all sub-pixels which are on the mask border.
        """
        self.image = image
        self.sub = sub

    @classmethod
    def from_mask_and_sub_grid_size(cls, mask, sub_grid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        """Setup the *ImagingGridBorders* from a mask, sub-grid size and polynomial degree.

        Parameters
        -----------
        mask : Mask
            The mask the padded border (sub-)pixel index's are computed from.
        sub_grid_size : int
            The size of a sub-pixels sub-grid (sub_grid_size x sub_grid_size).
        polynomial_degree : int
            The degree of the polynomial that is used to fit the border when relocating pixels outside the border to \
            its edge.
        centre : (float, float)
            The centre of the border, which can be shifted relative to its coordinates.
        """
        image_border = ImageGridBorder.from_mask(mask, polynomial_degree, centre)
        sub_border = SubGridBorder.from_mask(mask, sub_grid_size, polynomial_degree, centre)
        return ImagingGridBorders(image_border, sub_border)

    def relocated_grids_from_grids(self, grids):
        """Determine a set of relocated grids from an input set of grids, by relocating their pixels based on the \
        border.

        The blurring-grid does not have its coordinates relocated, as it is only used for computing analytic \
        light-profiles and not inversion-grids.

        Parameters
        -----------
        grids : ImagingGrids
            The imaging-grids (_data, sub) which have their coordinates relocated.
        """
        return ImagingGrids(image=self.image.relocated_grid_from_grid(grids.image),
                            sub=self.sub.relocated_grid_from_grid(grids.sub),
                            blurring=None)


class ImageGridBorder(np.ndarray):

    def __new__(cls, arr, polynomial_degree=3, centre=(0.0, 0.0), *args, **kwargs):
        """The borders of an _data-grid, containing the pixel-index's of all padded pixels that are on the \
        mask's border (e.g. they are next to a *True* value in at least one of the surrounding 8 pixels).

        A polynomial is fitted to the (x,y) coordinates of the border's pixels. This allows us to relocate \
        demagnified pixel's in a grid to its border, so that they do not disrupt an adaptive inversion.

        Parameters
        -----------
        arr : ndarray
            A 1D array of the integer indexes of an *ImageGrid*'s border pixels.
        polynomial_degree : int
            The degree of the polynomial that is used to fit the border when relocating pixels outside the border to \
            its edge.
        centre : (float, float)
            The centre of the border, which can be shifted relative to its coordinates.
        """
        border = arr.view(cls)
        border.polynomial_degree = polynomial_degree
        border.centre = centre
        return border

    @classmethod
    def from_mask(cls, mask, polynomial_degree=3, centre=(0.0, 0.0)):
        """Setup the *ImageGridBorder* from a mask.

        Parameters
        -----------
        mask : Mask
            The mask the padded border pixel index's are computed from.
        polynomial_degree : int
            The degree of the polynomial that is used to fit the border when relocating pixels outside the border to \
            its edge.
        centre : (float, float)
            The centre of the border, which can be shifted relative to its coordinates.
        """
        return cls(mask.border_pixels, polynomial_degree, centre)

    @property
    def total_pixels(self):
        return self.shape[0]

    def grid_to_radii(self, grid):
        """ Convert a grid of (x,y) arc-second coordinates to their circular radii values, based on the grid's centre.

        Parameters
        ----------
        grid : ImageGrid
            The (x, y) coordinates of each padded _data-pixel.
        """

        return np.sqrt(np.add(np.square(np.subtract(grid[:, 0], self.centre[0])),
                              np.square(np.subtract(grid[:, 1], self.centre[1]))))

    def grid_to_thetas(self, grid):
        """
        Converted a grid of (x,y) arc-second coordinates to the angle in degrees between their location and the \
        positive x-axis counter-clockwise.

        Parameters
        ----------
        grid : ImageGrid
            The (x, y) coordinates of each padded _data-pixel.
        """
        shifted_grid = np.subtract(grid, self.centre)
        theta_from_x = np.degrees(np.arctan2(shifted_grid[:, 1], shifted_grid[:, 0]))
        theta_from_x[theta_from_x < 0.0] += 360.
        return theta_from_x

    def polynomial_fit_to_border(self, grid):
        """Fit the border's radial coordinates with a polynomial.

        Parameters
        ----------
        grid : ImageGrid
            The (x, y) coordinates of each padded _data-pixel.
        """

        border_grid = grid[self]

        return np.polyfit(self.grid_to_thetas(border_grid), self.grid_to_radii(border_grid), self.polynomial_degree)

    def move_factors_from_grid(self, grid):
        """ Compute the move-factor of every (x,y) arc-second coordinate on the *ImageGrid*.

        A move-factor defines how far a coordinate must be moved towards the border's centre in order to lie on it. \
        If a pixel is already within the border, the move-factor 1.0 (i.e. no movement).

        Parameters
        ----------
        grid : ImageGrid
            The (x, y) coordinates of each padded _data-pixel.
        """
        grid_thetas = self.grid_to_thetas(grid)
        grid_radii = self.grid_to_radii(grid)
        poly = self.polynomial_fit_to_border(grid)

        with np.errstate(divide='ignore'):
            move_factors = np.divide(np.polyval(poly, grid_thetas), grid_radii)
        move_factors[move_factors > 1.0] = 1.0

        return move_factors

    def relocated_grid_from_grid(self, grid):
        """Compute the relocated (x,y) grid coordinates of every (x,y) arc-second coordinate on an *ImageGrid*.

        Parameters
        -----------
        grid : ImageGrid
            The (x, y) coordinates of each padded _data-pixel.
        """
        move_factors = self.move_factors_from_grid(grid)
        return np.multiply(grid, move_factors[:, None])

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(ImageGridBorder, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(ImageGridBorder, self).__setstate__(state[0:-1])


class SubGridBorder(ImageGridBorder):

    @classmethod
    def from_mask(cls, mask, sub_grid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        """Setup the *SubGridBorder* from a mask.

        Parameters
        -----------
        mask : Mask
            The mask the padded border pixel index's are computed from.
        sub_grid_size : int
            The size of a sub-pixels sub-grid (sub_grid_size x sub_grid_size).
        polynomial_degree : int
            The degree of the polynomial that is used to fit the border when relocating pixels outside the border to \
            its edge.
        centre : (float, float)
            The centre of the border, which can be shifted relative to its coordinates.
        """
        return cls(mask.border_sub_pixels(sub_grid_size), polynomial_degree, centre)



import inspect
import logging
from functools import wraps

import os
import numba
import numpy as np
from astropy.io import fits

from autolens import exc

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


class Memoizer(object):

    def __init__(self):
        """
        Class to store the results of a function given a set of inputs.
        """
        self.results = {}
        self.calls = 0
        self.arg_names = None

    def __call__(self, func):
        """
        Memoize decorator. Any time a function is called that a memoizer has been attached to its results are stored in
        the results dictionary or retrieved from the dictionary if the function has already been called with those
        arguments.

        Note that the same memoizer persists over all instances of a class. Any state for a given instance that is not
        given in the representation of that instance will be ignored. That is, it is possible that the memoizer will
        give incorrect results if instance state does not affect __str__ but does affect the value returned by the
        memoized method.

        Parameters
        ----------
        func: function
            A function for which results should be memoized

        Returns
        -------
        decorated: function
            A function that memoizes results
        """
        if self.arg_names is not None:
            raise AssertionError("Instantiate a new Memoizer for each function")
        self.arg_names = inspect.getfullargspec(func).args

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = ", ".join(
                ["('{}', {})".format(arg_name, arg) for arg_name, arg in
                 list(zip(self.arg_names, args)) + [(k, v) for k, v in kwargs.items()]])
            if key not in self.results:
                self.calls += 1
            self.results[key] = func(*args, **kwargs)
            return self.results[key]

        return wrapper


@numba.jit(nopython=True, cache=True)
def total_image_pixels_from_mask(mask):
    """Compute the total number of unmasked _data pixels in a mask."""

    total_image_pixels = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                total_image_pixels += 1

    return total_image_pixels


@numba.jit(nopython=True, cache=True)
def total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size):
    """Compute the total number of sub-pixels in unmasked _data pixels in a mask."""
    return total_image_pixels_from_mask(mask) * sub_grid_size ** 2


@numba.jit(nopython=True, cache=True)
def total_border_pixels_from_mask(mask):
    """Compute the total number of border-pixels in a mask."""

    border_pixel_total = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                if mask[y + 1, x] or mask[y - 1, x] or mask[y, x + 1] or mask[y, x - 1] or \
                        mask[y + 1, x + 1] or mask[y + 1, x - 1] or mask[y - 1, x + 1] or mask[y - 1, x - 1]:
                    border_pixel_total += 1

    return border_pixel_total


@numba.jit(nopython=True, cache=True)
def image_grid_2d_from_shape_and_pixel_scales(shape, pixel_scales):
    """
    Computes the (x,y) arc second coordinates of every pixel in an _data of shape (rows, columns).

    Coordinates are defined from the top-left corner, such that the first pixel at location [0, 0] has negative x \
    and y values in arc seconds.
    """

    grid_2d = np.zeros((shape[0], shape[1], 2))

    y_cen = float(shape[0] - 1) / 2
    x_cen = float(shape[1] - 1) / 2

    for y in range(shape[0]):
        for x in range(shape[1]):

            grid_2d[y, x, 0] = -(y - y_cen) * pixel_scales[0]
            grid_2d[y, x, 1] = (x - x_cen) * pixel_scales[1]

    return grid_2d


@numba.jit(nopython=True, cache=True)
def image_grid_1d_from_shape_and_pixel_scales(shape, pixel_scales):
    """
    Computes the (x,y) arc second coordinates of every pixel in an _data of shape (rows, columns).

    Coordinates are defined from the top-left corner, such that the first pixel at location [0, 0] has negative x \
    and y values in arc seconds.
    """

    grid_1d = np.zeros((shape[0]*shape[1], 2))

    y_cen = float(shape[0] - 1) / 2
    x_cen = float(shape[1] - 1) / 2

    i=0
    for y in range(shape[0]):
        for x in range(shape[1]):

            grid_1d[i, 0] = -(y - y_cen) * pixel_scales[0]
            grid_1d[i, 1] = (x - x_cen) * pixel_scales[1]
            i += 1

    return grid_1d

@numba.jit(nopython=True, cache=True)
def image_grid_1d_masked_from_mask_and_pixel_scales(mask, pixel_scales):
    """Compute a 1D grid of (x,y) coordinates, using the center of every unmasked pixel."""

    grid_2d = image_grid_2d_from_shape_and_pixel_scales(mask.shape, pixel_scales)

    total_image_pixels = total_image_pixels_from_mask(mask)
    image_grid = np.zeros(shape=(total_image_pixels, 2))
    pixel_count = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                image_grid[pixel_count, :] = grid_2d[y, x]
                pixel_count += 1

    return image_grid

@numba.jit(nopython=True, cache=True)
def sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask, pixel_scales, sub_grid_size):
    """Compute a 1D grid of (x,y) sub-pixel coordinates, using the sub-pixel centers of every unmasked pixel."""

    total_sub_pixels = total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size)

    sub_grid = np.zeros(shape=(total_sub_pixels, 2))

    y_cen = float(mask.shape[0] - 1) / 2
    x_cen = float(mask.shape[1] - 1) / 2

    sub_index = 0

    y_sub_half = pixel_scales[0] / 2
    y_sub_step = pixel_scales[0] / (sub_grid_size + 1)

    x_sub_half = pixel_scales[1] / 2
    x_sub_step = pixel_scales[1] / (sub_grid_size + 1)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            if not mask[y, x]:

                y_arcsec = (y - y_cen) * pixel_scales[0]
                x_arcsec = (x - x_cen) * pixel_scales[1]

                for y1 in range(sub_grid_size):
                    for x1 in range(sub_grid_size):

                        sub_grid[sub_index, 0] = -(y_arcsec - y_sub_half + (y1 + 1) * y_sub_step)
                        sub_grid[sub_index, 1] = x_arcsec - x_sub_half + (x1 + 1) * x_sub_step
                        sub_index += 1

    return sub_grid

@numba.jit(nopython=True, cache=True)
def grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds, shape, pixel_scales):
    """ Converts a grid in coordinates of pixels to a grid in arc seconds.

    The pixel coordinate origin is at the top left corner of an image, whilst the arc-second coordinate origin \
    is at the centre start with negative x and y values from the top-left.

    This means that the top-left pixel coordinates, [0, 0], will give negative arc second coordinates.

    Parameters
    ----------
    grid_pixels : ndarray
        The grid of (x,y) coordinates in units of pixels
    """

    grid_pixels = np.zeros((grid_arc_seconds.shape[0], 2))

    y_cen = float(shape[0] - 1) / 2
    x_cen = float(shape[1] - 1) / 2

    for i in range(grid_arc_seconds.shape[0]):

        grid_pixels[i, 0] =(-grid_arc_seconds[i,0] / pixel_scales[0])  + y_cen + 0.5
        grid_pixels[i, 1] =(grid_arc_seconds[i,1] / pixel_scales[1])  + x_cen + 0.5

    return grid_pixels

@numba.jit(nopython=True, cache=True)
def grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds, shape, pixel_scales):
    """ Converts a grid in coordinates of pixels to a grid in arc seconds.

    The pixel coordinate origin is at the top left corner of an image, whilst the arc-second coordinate origin \
    is at the centre start with negative x and y values from the top-left.

    This means that the top-left pixel coordinates, [0, 0], will give negative arc second coordinates.

    Parameters
    ----------
    grid_pixels : ndarray
        The grid of (x,y) coordinates in units of pixels
    """

    grid_pixels = np.zeros((grid_arc_seconds.shape[0], 2))

    y_cen = float(shape[0] - 1) / 2
    x_cen = float(shape[1] - 1) / 2

    for i in range(grid_arc_seconds.shape[0]):

        grid_pixels[i, 0] = int((-grid_arc_seconds[i,0] / pixel_scales[0])  + y_cen + 0.5)
        grid_pixels[i, 1] = int((grid_arc_seconds[i,1] / pixel_scales[1])  + x_cen + 0.5)

    return grid_pixels

@numba.jit(nopython=True, cache=True)
def grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels, shape, pixel_scales):
    """ Converts a grid in coordinates of pixels to a grid in arc seconds.

    The pixel coordinate origin is at the top left corner of an image, whilst the arc-second coordinate origin \
    is at the centre start with negative x and y values from the top-left.

    This means that the top-left pixel coordinates, [0, 0], will give negative arc second coordinates.

    Parameters
    ----------
    grid_pixels : ndarray
        The grid of (x,y) coordinates in units of pixels
    """

    grid_arc_seconds = np.zeros((grid_pixels.shape[0], 2))

    y_cen = float(shape[0] - 1) / 2
    x_cen = float(shape[1] - 1) / 2

    for i in range(grid_arc_seconds.shape[0]):

        grid_arc_seconds[i, 0] = -(grid_pixels[i,0] - y_cen - 0.5) * pixel_scales[0]
        grid_arc_seconds[i, 1] = (grid_pixels[i,1] - x_cen - 0.5) * pixel_scales[1]

    return grid_arc_seconds

@numba.jit(nopython=True, cache=True)
def grid_to_pixel_from_mask(mask):
    """Compute a 1D array that maps every unmasked pixel to its corresponding 2d pixel using its (x,y) pixel indexes.

    For howtolens if pixel [2,5] corresponds to the second pixel on the 1D array, grid_to_pixel[1] = [2,5]"""

    total_image_pixels = total_image_pixels_from_mask(mask)
    grid_to_pixel = np.zeros(shape=(total_image_pixels, 2))
    pixel_count = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                grid_to_pixel[pixel_count, :] = y, x
                pixel_count += 1

    return grid_to_pixel

@numba.jit(nopython=True, cache=True)
def sub_to_image_from_mask(mask, sub_grid_size):
    """Compute a 1D array that maps every unmasked pixel's sub-pixel to its corresponding 1d _data-pixel.

    For howtolens, if sub-pixel 8 is in _data-pixel 1, sub_to_image[7] = 1."""

    total_sub_pixels = total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size)

    sub_to_image = np.zeros(shape=total_sub_pixels)
    image_index = 0
    sub_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range(sub_grid_size):
                    for x1 in range(sub_grid_size):
                        sub_to_image[sub_index] = image_index
                        sub_index += 1

                image_index += 1

    return sub_to_image

@numba.jit(nopython=True, cache=True)
def border_pixels_from_mask(mask):
    """Compute a 1D array listing all border pixel indexes in the mask. A border pixel is a pixel which is not fully \
    surrounding by False mask values i.e. it is on an edge."""

    border_pixel_total = total_border_pixels_from_mask(mask)

    border_pixels = np.zeros(border_pixel_total)
    border_index = 0
    image_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                if mask[y + 1, x] or mask[y - 1, x] or mask[y, x + 1] or mask[y, x - 1] or \
                        mask[y + 1, x + 1] or mask[y + 1, x - 1] or mask[y - 1, x + 1] or mask[y - 1, x - 1]:
                    border_pixels[border_index] = image_index
                    border_index += 1

                image_index += 1

    return border_pixels


@numba.jit(nopython=True, cache=True)
def border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask, pixel_scales, sub_grid_size):
    """Compute a 1D array listing all sub-pixel border pixel indexes in the mask. A border sub-pixel is a sub-pixel \
    whose _data pixel is not fully surrounded by False mask values and it is closest to the edge."""
    border_pixel_total = total_border_pixels_from_mask(mask)
    border_sub_pixels = np.zeros(border_pixel_total)

    image_index = 0

    y_cen = float(mask.shape[0] - 1) / 2
    x_cen = float(mask.shape[1] - 1) / 2

    y_sub_half = pixel_scales[0] / 2
    y_sub_step = pixel_scales[0] / (sub_grid_size + 1)

    x_sub_half = pixel_scales[1] / 2
    x_sub_step = pixel_scales[1] / (sub_grid_size + 1)

    border_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                if mask[y + 1, x] or mask[y - 1, x] or mask[y, x + 1] or mask[y, x - 1] or \
                        mask[y + 1, x + 1] or mask[y + 1, x - 1] or mask[y - 1, x + 1] or mask[y - 1, x - 1]:

                    y_arcsec = (y - y_cen) * pixel_scales[0]
                    x_arcsec = (x - x_cen) * pixel_scales[1]

                    sub_grid = np.zeros((sub_grid_size ** 2, 2))
                    sub_index = 0

                    for y1 in range(sub_grid_size):
                        for x1 in range(sub_grid_size):
                            sub_grid[sub_index, 0] = y_arcsec - y_sub_half + (y1 + 1) * y_sub_step
                            sub_grid[sub_index, 1] = x_arcsec - x_sub_half + (x1 + 1) * x_sub_step
                            sub_index += 1

                    sub_grid_radii = np.add(np.square(sub_grid[:, 0]), np.square(sub_grid[:, 1]))
                    border_sub_index = image_index * (sub_grid_size ** 2) + np.argmax(sub_grid_radii)
                    border_sub_pixels[border_index] = border_sub_index
                    border_index += 1

                image_index += 1

    return border_sub_pixels


@numba.jit(nopython=True, cache=True)
def mask_circular_from_shape_pixel_scale_and_radius(shape, pixel_scale, radius_arcsec, centre=(0.0, 0.0)):
    """Compute a circular mask from an input mask radius and _data shape."""

    mask = np.full(shape, True)

    y_cen = (float(mask.shape[0] - 1) / 2) - (centre[0] / pixel_scale)
    x_cen = (float(mask.shape[1] - 1) / 2) + (centre[1] / pixel_scale)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - y_cen) * pixel_scale
            x_arcsec = (x - x_cen) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if r_arcsec <= radius_arcsec:
                mask[y, x] = False

    return mask


@numba.jit(nopython=True, cache=True)
def mask_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec,
                                                  centre=(0.0, 0.0)):
    """Compute an annular mask from an input inner and outer mask radius and _data shape."""

    mask = np.full(shape, True)

    y_cen = (float(mask.shape[0] - 1) / 2) - (centre[0] / pixel_scale)
    x_cen = (float(mask.shape[1] - 1) / 2) + (centre[1] / pixel_scale)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - y_cen) * pixel_scale
            x_arcsec = (x - x_cen) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if outer_radius_arcsec >= r_arcsec >= inner_radius_arcsec:
                mask[y, x] = False

    return mask

@numba.jit(nopython=True, cache=True)
def mask_anti_annular_from_shape_pixel_scale_and_radii(shape, pixel_scale, inner_radius_arcsec, outer_radius_arcsec,
                                                       outer_radius_2_arcsec, centre=(0.0, 0.0)):
    """Compute an annular mask from an input inner and outer mask radius and _data shape."""

    mask = np.full(shape, True)

    y_cen = (float(mask.shape[1] - 1) / 2) - (centre[1] / pixel_scale)
    x_cen = (float(mask.shape[0] - 1) / 2) + (centre[0] / pixel_scale)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):

            y_arcsec = (y - y_cen) * pixel_scale
            x_arcsec = (x - x_cen) * pixel_scale

            r_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

            if  inner_radius_arcsec >= r_arcsec or outer_radius_2_arcsec >= r_arcsec >= outer_radius_arcsec:
                mask[y, x] = False

    return mask

@numba.jit(nopython=True, cache=True)
def mask_blurring_from_mask_and_psf_shape(mask, psf_shape):
    """Compute a blurring mask from an input mask and psf shape.

    The blurring mask corresponds to all pixels which are outside of the mask but will have a fraction of their \
    light blur into the masked region due to PSF convolution."""

    blurring_mask = np.full(mask.shape, True)

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range((-psf_shape[0] + 1) // 2, (psf_shape[0] + 1) // 2):
                    for x1 in range((-psf_shape[1] + 1) // 2, (psf_shape[1] + 1) // 2):
                        if 0 <= x + x1 <= mask.shape[1] - 1 and 0 <= y + y1 <= mask.shape[0] - 1:
                            if mask[y + y1, x + x1]:
                                blurring_mask[y + y1, x + x1] = False
                        else:
                            raise exc.MaskException(
                                "setup_blurring_mask extends beyond the sub_grid_size of the mask - pad the "
                                "masked_image before masking")

    return blurring_mask


@numba.jit(nopython=True, cache=True)
def map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d):
    """For a given 2D array and mask, mappers all unmasked pixels to a 1D array."""

    total_image_pixels = total_image_pixels_from_mask(mask)

    array_1d = np.zeros(shape=total_image_pixels)
    index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                array_1d[index] = array_2d[y, x]
                index += 1

    return array_1d


@numba.jit(nopython=True, cache=True)
def map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d, shape, one_to_two):
    """For a masked 1D array, mappers it to a 2D array using the mappings from 1D to 2D (one_to_two)."""

    array_2d = np.zeros(shape)

    for index in range(len(one_to_two)):
        array_2d[one_to_two[index, 0], one_to_two[index, 1]] = array_1d[index]

    return array_2d


@numba.jit(nopython=True, cache=True)
def map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d, shape):
    """Map a 1D array where every value in its original 2D array was unmasked, to this original 2D array."""

    array_2d = np.zeros(shape)

    index = 0
    for y in range(shape[0]):
        for x in range(shape[1]):
            array_2d[y, x] = array_1d[index]
            index += 1

    return array_2d


def pad_array_2d_around_centre(array_2d, new_shape):
    """
    Trim the data_vector array to a new sub_grid_size around its central pixel.

    NOTE: The centre of the array cannot be shifted. Therefore, even arrays must be trimmed to even arrays \
    (e.g. 8x8 -> 4x4) and odd to odd (e.g. 5x5 -> 3x3).

    Parameters
    ----------
    array_2d
    new_shape : (int, int)
        The (x,y) new pixel dimension of the trimmed data_vector-array.
    """

    if new_shape[0] < array_2d.shape[0]:
        raise ValueError(
            'grids.Grid2d.pad_data - You have specified a new x_size smaller than the input 2d array')
    elif new_shape[1] < array_2d.shape[1]:
        raise ValueError(
            'grids.Grid2d.pad_data - You have specified a new y_size bigger than the input 2d array')

    if array_2d.shape[0] % 2 == 0 and not new_shape[0] % 2 == 0:
        raise ValueError('You cannot pad an array from even shape to odd shape - change new_shape to even')

    if array_2d.shape[1] % 2 == 0 and not new_shape[1] % 2 == 0:
        raise ValueError('You cannot pad an array from even shape to odd shape - change new_shape to even')

    if not array_2d.shape[0] % 2 == 0 and new_shape[0] % 2 == 0:
        raise ValueError('You cannot pad an array from odd shape to even shape - change new_shape to odd')

    if not array_2d.shape[1] % 2 == 0 and new_shape[1] % 2 == 0:
        raise ValueError('You cannot pad an array from odd shape to even shape - change new_shape to odd')

    y_pad = int((new_shape[0] - array_2d.shape[0]) / 2)
    x_pad = int((new_shape[1] - array_2d.shape[1]) / 2)

    return np.pad(array=array_2d, pad_width=(y_pad, x_pad), mode='constant')


def trim_array_2d_around_centre(array_2d, new_shape):
    """
    Trim the data_vector array to a new sub_grid_size around its central pixel.

    NOTE: The centre of the array cannot be shifted. Therefore, even arrays must be trimmed to even arrays \
    (e.g. 8x8 -> 4x4) and odd to odd (e.g. 5x5 -> 3x3).

    Parameters
    ----------
    array_2d
    new_shape : (int, int)
        The (x,y) new pixel dimension of the trimmed data_vector-array.
    """

    if new_shape[0] > array_2d.shape[0]:
        raise ValueError(
            'grids.Grid2d.trim_data - You have specified a new x_size bigger than the data_vector array')
    elif new_shape[1] > array_2d.shape[1]:
        raise ValueError(
            'grids.Grid2d.trim_data - You have specified a new y_size bigger than the data_vector array')

    if array_2d.shape[0] % 2 == 0 and not new_shape[0] % 2 == 0:
        raise ValueError('You cannot trim an array from even shape to odd shape - change new_shape to even')

    if array_2d.shape[1] % 2 == 0 and not new_shape[1] % 2 == 0:
        raise ValueError('You cannot trim an array from even shape to odd shape - change new_shape to even')

    if not array_2d.shape[0] % 2 == 0 and new_shape[0] % 2 == 0:
        raise ValueError('You cannot trim an array from odd shape to even shape - change new_shape to odd')

    if not array_2d.shape[1] % 2 == 0 and new_shape[1] % 2 == 0:
        raise ValueError('You cannot trim an array from odd shape to even shape - change new_shape to odd')

    y_trim = int((array_2d.shape[0] - new_shape[0]) / 2)
    x_trim = int((array_2d.shape[1] - new_shape[1]) / 2)

    array = array_2d[y_trim:array_2d.shape[0] - y_trim, x_trim:array_2d.shape[1] - x_trim]

    return array


def trim_array_2d_around_region(array_2d, y0, y1, x0, x1):
    """
    Trim the data_vector array to a new sub_grid_size around its central pixel.

    NOTE: The centre of the array cannot be shifted. Therefore, even arrays must be trimmed to even arrays \
    (e.g. 8x8 -> 4x4) and odd to odd (e.g. 5x5 -> 3x3).

    Parameters
    ----------
    array_2d
    new_shape : (int, int)
        The (x,y) new pixel dimension of the trimmed data_vector-array.
    """

    if y1 > array_2d.shape[1]:
        raise ValueError(
            'grids.Grid2d.trim_data - You have specified a new y_size bigger than the data_vector array')
    elif x1 > array_2d.shape[0]:
        raise ValueError(
            'grids.Grid2d.trim_data - You have specified a new x_size bigger than the data_vector array')

    return array_2d[y0:y1, x0:x1]


def numpy_array_to_fits(array, path, overwrite=False):

    if overwrite and os.path.exists(path):
        os.remove(path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(np.flipud(array), new_hdr)
    hdu.writeto(path)


def numpy_array_from_fits(path, hdu):
    hdu_list = fits.open(path)
    return np.flipud(np.array(hdu_list[hdu].data))


def compute_variances_from_noise(noise):
    """The variances are the signal_to_noise_ratio (standard deviations) squared."""
    return np.square(noise)



import scipy
from analysis import galaxy
from analysis import ray_tracing
from profiling import profiling_data
from profiling import tools

from profiles import mass_profiles


class Voronoi(object):

    def __init__(self, pixels=100, regularization_coefficients=(1.0,)):
        """
        Abstract base class for a Voronoi inversion, which represents pixels as a set of centers where \
        all of the nearest-neighbor pix-grid (i.e. traced masked_image-pixels) are mapped to them.

        This forms a Voronoi grid pix-plane, the properties of which are used for fast calculations, defining the \
        regularization_matrix matrix and visualization.

        Parameters
        ----------
        pixels : int
            The number of pixels in the inversion.
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """

        super(Voronoi, self).__init__(pixels, regularization_coefficients)

    @staticmethod
    def voronoi_from_cluster_grid(cluster_grid):
        """Compute the Voronoi grid of the inversion, using the pixel centers.

        Parameters
        ----------
        cluster_grid : ndarray
            The x and y image_grid to derive the Voronoi grid_coords.
        """
        return scipy.spatial.Voronoi(cluster_grid, qhull_options='Qbb Qc Qx Qm')


sub_grid_size = 4

pix = Voronoi(pixels=200)

sie = mass_profiles.EllipticalIsothermal(centre=(0.010, 0.032), einstein_radius=1.47, axis_ratio=0.849, phi=73.6)
shear = mass_profiles.ExternalShear(magnitude=0.0663, phi=160.5)

lens_galaxy = galaxy.Galaxy(mass_profile_0=sie, mass_profile_1=shear)

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

lsst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[], image_plane_grids=lsst.grids)
euclid_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[], image_plane_grids=euclid.grids)
hst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[], image_plane_grids=hst.grids)
hst_up_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[], image_plane_grids=hst_up.grids)
ao_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[], image_plane_grids=ao.grids)


@tools.tick_toc_x1
def lsst_solution():
    pix.voronoi_from_cluster_grid()


@tools.tick_toc_x1
def euclid_solution():
    pix.mapping_matrix_from_sub_to_pix(sub_to_pix=euclid_sub_to_pix, grids=euclid.grids)


@tools.tick_toc_x1
def hst_solution():
    pix.mapping_matrix_from_sub_to_pix(sub_to_pix=hst_sub_to_pix, grids=hst.grids)


@tools.tick_toc_x1
def hst_up_solution():
    pix.mapping_matrix_from_sub_to_pix(sub_to_pix=hst_up_sub_to_pix, grids=hst_up.grids)


@tools.tick_toc_x1
def ao_solution():
    pix.mapping_matrix_from_sub_to_pix(sub_to_pix=ao_sub_to_pix, grids=ao.grids)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import numpy as np
import pytest

from autolens import exc
from autolens.imaging import image
from autolens.imaging import mask
from autolens.inversion import convolution

class TestConvolveMappingMatrix(object):

    def test__asymetric_convolver__matrix_blurred_correctly(self):
        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0],
                                      [0.4, 0.2, 0.3],
                                      [0, 0.1, 0]])

        convolver = convolution.ConvolverMappingMatrix(mask=mask, psf=asymmetric_kernel)

        mapping = np.array([[0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 1, 0],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                            [0, 0, 0],
                            [1, 0, 0],
                            [0, 0, 1],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0]])

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert (blurred_mapping == np.array([[0, 0, 0],
                                             [0, 0, 0],
                                             [0, 0, 0],
                                             [0, 0, 0],
                                             [0, 0, 0],
                                             [0, 0, 0],
                                             [0, 0.4, 0],
                                             [0, 0.2, 0],
                                             [0.4, 0, 0],
                                             [0.2, 0, 0.4],
                                             [0.3, 0, 0.2],
                                             [0, 0.1, 0.3],
                                             [0, 0, 0],
                                             [0.1, 0, 0],
                                             [0, 0, 0.1],
                                             [0, 0, 0]])).all()

    def test__asymetric_convolver__multiple_overlapping_blurred_entires_in_matrix(self):
        shape = (4, 4)
        mask = np.full(shape, False)

        asymmetric_kernel = np.array([[0, 0.0, 0],
                                      [0.4, 0.2, 0.3],
                                      [0, 0.1, 0]])

        convolver = convolution.ConvolverMappingMatrix(mask=mask, psf=asymmetric_kernel)

        mapping = np.array([[0, 1, 0],
                            [0, 1, 0],
                            [0, 1, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 1, 0],  # The 0.3 should be 'chopped' from this pixel as it is on the right-most edge
                            [1, 0, 0],
                            [1, 0, 0],
                            [0, 0, 1],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0]])

        blurred_mapping = convolver.convolve_mapping_matrix(mapping)

        assert blurred_mapping == pytest.approx(np.array([[0, 0.6, 0],
                                                          [0, 0.9, 0],
                                                          [0, 0.5, 0],
                                                          [0, 0.3, 0],
                                                          [0, 0.1, 0],
                                                          [0, 0.1, 0],
                                                          [0, 0.5, 0],
                                                          [0, 0.2, 0],
                                                          [0.6, 0, 0],
                                                          [0.5, 0, 0.4],
                                                          [0.3, 0, 0.2],
                                                          [0, 0.1, 0.3],
                                                          [0.1, 0, 0],
                                                          [0.1, 0, 0],
                                                          [0, 0, 0.1],
                                                          [0, 0, 0]]), 1e-4)


import colorama

from autolens import conf
from autolens import runners
from autolens.commands.base import Base, prepend_working_directory


def color(text, fore):
    """
    Apply a color to some text.

    Parameters
    ----------
    text: str
        The original text
    fore: colorama.ansi.AnsiFore
        The color to be applied to the text

    Returns
    -------
    text: str
        Colored text
    """
    return "{}{}{}".format(fore, text, colorama.Fore.RESET)


def blue(text):
    """
    Make text blue
    """
    return color(text, colorama.Fore.BLUE)


def red(text):
    """
    Make text red
    """
    return color(text, colorama.Fore.RED)


class Pipeline(Base):

    def run(self):
        name = self.options['<name>']
        conf.instance = conf.Config(self.config_path, self.output_path)
        try:
            if self.options['--info']:
                tup = runners.pipeline_dict[name]
                print()
                pl = tup.make()
                print(red(name))
                print(tup.doc)
                print()
                print(red("Phases"))
                print("\n".join(["{}\n   {}".format(phase.__class__.__name__, blue(phase.doc)) for phase in pl.phases]))
                return
            if name is not None:
                if name == "test":
                    self.run_pipeline(runners.TestPipeline())
                    return
                self.run_pipeline(runners.pipeline_dict[name].make())
                return
        except KeyError:
            print("Pipeline '{}' does not exist.\n".format(name))

        print_pipelines()

    def run_pipeline(self, pl):
        from autolens.imaging import image as im
        if self.is_using_hdu:
            image = im.load_imaging_from_fits(self.data_path, self.image_hdu, self.noise_hdu, self.psf_hdu,
                                              self.pixel_scale)
        else:
            image = im.load_imaging_from_path(self.image_path, self.noise_path, self.psf_path,
                                              pixel_scale=self.pixel_scale)
        pl.run(image)

    @property
    def is_using_hdu(self):
        """
        Returns
        -------
        is_using_hdu: bool
            True iff --data option is set. --data is the path to a file with multiple data layers accessible by setting
            hdus.
        """
        return self.options["--data"] is not None

    @property
    def image_hdu(self):
        """
        Returns
        -------
        str: image_hdu
            The hdu of the image data in the data file
        """
        return int(self.options["--image-hdu"])

    @property
    def noise_hdu(self):
        """
        Returns
        -------
        str: noise_hdu
            The hdu of the noise data in the data file
        """
        return int(self.options["--noise-hdu"])

    @property
    def psf_hdu(self):
        """
        Returns
        -------
        str: psf_hdu
            The hdu of the psf data in the data file
        """
        return int(self.options["--psf-hdu"])

    @property
    @prepend_working_directory
    def image_path(self):
        """
        Get the relative or absolute path to the input _data. If the path does not begin with '/' then the current
        working directory will be prepended.

        Returns
        -------
        str: path
            The path to the image
        """
        return self.options['--image']

    @property
    @prepend_working_directory
    def data_path(self):
        """
        Get the relative or absolute path to the input data. Input data includes _data, noise and psf with different
        hdu values input by the user. If the path does not begin with '/' then the current working directory will be
        prepended.

        Returns
        -------
        str: path
            The path to the data
        """
        return self.options['--data']

    @property
    @prepend_working_directory
    def noise_path(self):
        """
        Get the relative or absolute path to the input noise. If the path does not begin with '/' then the current
        working directory will be prepended.

        Returns
        -------
        str: path
            The path to the noise
        """
        return self.options['--noise']

    @property
    @prepend_working_directory
    def psf_path(self):
        """
        Get the relative or absolute path to the input psf. If the path does not begin with '/' then the current
        working directory will be prepended.

        Returns
        -------
        str: path
            The path to the psf folder or psf.
        """
        return self.options['--psf']

    @property
    def pixel_scale(self):
        """
        Returns
        -------
        pixel_scales: float
            The size of a single pixel, in arc seconds, as input by the user
        """
        return float(self.options['--pixel-scale'])

    @property
    def config_path(self):
        """
        Returns
        -------
        config_path: str
            The path to the configuration folder. Defaults to 'config' in the current working directory.
        """
        if '--config' in self.options:
            config_path = self.options['--config']
        else:
            config_path = 'config'
        return config_path

    @property
    @prepend_working_directory
    def output_path(self):
        """
        Returns
        -------
        output_path: str
            The path to the configuration folder. Defaults to 'output' in the current working directory.
        """
        return self.options['--output']


def print_pipelines():
    """
    Prints a list of available runners taken from the pipeline dictionary.
    """
    print("Available Pipelines:\n")
    print(
        "\n".join(
            ["{}\n  {}".format(key, blue(value.short_doc)) for
             key, value
             in
             runners.pipeline_dict.items()]))



import numba
import numpy as np
import pytest
from profiling import profiling_data
from profiling import tools

from autolens import exc


class Pixelization(object):

    def __init__(self, pixels, regularization_coefficients=(1.0,), pix_signal_scale=1.0):
        """
        Abstract base class for a inversion, which discretizes a set of masked_image and sub grid grid into \
        pixels. These pixels then incorrect_fit a  data_vector-set using a linear inversion, where their regularization_matrix matrix
        enforces smoothness between pixel values.

        A number of 1D and 2D arrays are used to represent mappings betwen masked_image, sub, pix, and cluster pixels. The \
        nomenclature here follows grid_to_grid, such that it maps the index of a value on one grid to another. For \
        howtolens:

        - pix_to_image[2] = 5 tells us that the 3rd inversion-pixel maps to the 6th masked_image-pixel.
        - sub_to_pixelization[4,2] = 2 tells us that the 5th sub-pixel maps to the 3rd inversion-pixel.

        Parameters
        ----------
        pixels : int
            The number of pixels in the inversion.
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        pix_signal_scale : float
            A hyper-parameter which scales the signal attributed to each pixel, used for weighted regularization_matrix.
        """
        self.pixels = pixels
        self.regularization_coefficients = regularization_coefficients
        self.pix_signal_scale = pix_signal_scale


class Rectangular(Pixelization):

    def __init__(self, shape=(3, 3), regularization_coefficients=(1.0,)):
        """A rectangular inversion where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an masked_image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular inversion must be at least dimensions 3x3')

        super(Rectangular, self).__init__(shape[0] * shape[1], regularization_coefficients)

        self.shape = shape

    class Geometry(object):

        def __init__(self, x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale):
            """The geometry of a rectangular grid, defining where the grids top-left, top-right, bottom-left and \
            bottom-right corners are in arc seconds. The arc-second size of each rectangular pixel is also computed.

            Parameters
            -----------

            """
            self.x_min = x_min
            self.x_max = x_max
            self.x_pixel_scale = x_pixel_scale
            self.y_min = y_min
            self.y_max = y_max
            self.y_pixel_scale = y_pixel_scale

        def arc_second_to_pixel_index_x(self, coordinate):
            return np.floor((coordinate - self.x_min) / self.x_pixel_scale)

        def arc_second_to_pixel_index_y(self, coordinate):
            return np.floor((coordinate - self.y_min) / self.y_pixel_scale)

    def geometry_from_pix_sub_grid(self, pix_sub_grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pix_grid grid \
        plus a small buffer.

        Parameters
        -----------
        pix_sub_grid : [[float, float]]
            The x and y pix grid (or sub-coordinaates) which are to be matched with their pixels.
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        x_min = np.min(pix_sub_grid[:, 0]) - buffer
        x_max = np.max(pix_sub_grid[:, 0]) + buffer
        y_min = np.min(pix_sub_grid[:, 1]) - buffer
        y_max = np.max(pix_sub_grid[:, 1]) + buffer
        x_pixel_scale = (x_max - x_min) / self.shape[0]
        y_pixel_scale = (y_max - y_min) / self.shape[1]

        return self.Geometry(x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale)

    def grid_to_pix_from_grid(self, grid, geometry):
        """Compute the mappings between a set of masked_image pixels (or sub-pixels) and pixels, using the masked_image's
        traced pix-plane grid (or sub-grid) and the uniform rectangular inversion's geometry.

        Parameters
        ----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        geometry : Geometry
            The rectangular pixel grid's geometry.
        """
        grid_to_pix = np.zeros(grid.shape[0], dtype='int')

        for index, pix_coordinate in enumerate(grid):
            x_pixel = geometry.x_arc_seconds_to_pixels(pix_coordinate[0])
            y_pixel = geometry.y_arc_seconds_to_pixels(pix_coordinate[1])

            grid_to_pix[index] = x_pixel * self.shape[1] + y_pixel

        return grid_to_pix

    def grid_to_pix_from_grid_jitted(self, grid, geometry):
        """Compute the mappings between a set of masked_image pixels (or sub-pixels) and pixels, using the masked_image's
        traced pix-plane grid (or sub-grid) and the uniform rectangular inversion's geometry.

        Parameters
        ----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        geometry : Geometry
            The rectangular pixel grid's geometry.
        """
        return self.grid_to_pix_jit(grid, geometry.x_min, geometry.x_pixel_scale, geometry.y_min,
                                    geometry.y_pixel_scale, self.shape[1]).astype(dtype='int')

    @staticmethod
    @numba.jit(nopython=True)
    def grid_to_pix_jit(grid, x_min, x_pixel_scale, y_min, y_pixel_scale, y_shape):

        grid_to_pix = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            x_pixel = np.floor((grid[i, 0] - x_min) / x_pixel_scale)
            y_pixel = np.floor((grid[i, 1] - y_min) / y_pixel_scale)

            grid_to_pix[i] = x_pixel * y_shape + y_pixel

        return grid_to_pix


sub_grid_size = 4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

pix_shape = (50, 50)

pix = Rectangular(pix_shape)

lsst_geometry = pix.geometry_from_pix_sub_grid(pix_sub_grid=lsst.grids.sub)
euclid_geometry = pix.geometry_from_pix_sub_grid(pix_sub_grid=euclid.grids.sub)
hst_geometry = pix.geometry_from_pix_sub_grid(pix_sub_grid=hst.grids.sub)
hst_up_geometry = pix.geometry_from_pix_sub_grid(pix_sub_grid=hst_up.grids.sub)
ao_geometry = pix.geometry_from_pix_sub_grid(pix_sub_grid=ao.grids.sub)

assert pix.grid_to_pix_from_grid(grid=lsst.grids.sub, geometry=lsst_geometry) == \
       pytest.approx(pix.grid_to_pix_from_grid_jitted(grid=lsst.grids.sub, geometry=lsst_geometry))
pix.grid_to_pix_from_grid_jitted(grid=lsst.grids.sub, geometry=lsst_geometry)
pix.grid_to_pix_from_grid_jitted(grid=euclid.grids.sub, geometry=euclid_geometry)
pix.grid_to_pix_from_grid_jitted(grid=hst.grids.sub, geometry=hst_geometry)
pix.grid_to_pix_from_grid_jitted(grid=hst_up.grids.sub, geometry=hst_up_geometry)
pix.grid_to_pix_from_grid_jitted(grid=ao.grids.sub, geometry=ao_geometry)


@tools.tick_toc_x1
def lsst_solution():
    pix.grid_to_pix_from_grid_jitted(grid=lsst.grids.sub, geometry=lsst_geometry)


@tools.tick_toc_x1
def euclid_solution():
    pix.grid_to_pix_from_grid_jitted(grid=euclid.grids.sub, geometry=euclid_geometry)


@tools.tick_toc_x1
def hst_solution():
    pix.grid_to_pix_from_grid_jitted(grid=hst.grids.sub, geometry=hst_geometry)


@tools.tick_toc_x1
def hst_up_solution():
    pix.grid_to_pix_from_grid_jitted(grid=hst_up.grids.sub, geometry=hst_up_geometry)


@tools.tick_toc_x1
def ao_solution():
    pix.grid_to_pix_from_grid_jitted(grid=ao.grids.sub, geometry=ao_geometry)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import numpy as np
import pytest

from autolens.imaging import imaging_util
from autolens.inversion import mappers
from autolens.inversion import pixelizations
from test.mock.mock_imaging import MockSubGridCoords, MockGridCollection


class MockGeometry(object):

    def __init__(self):

        self.pixel_scales = (1.0, 1.0)


def grid_to_pixel_pixels_via_nearest_neighbour(grid, pixel_centers):
    def compute_squared_separation(coordinate1, coordinate2):
        """Computes the squared separation of two image_grid (no square root for efficiency)"""
        return (coordinate1[0] - coordinate2[0]) ** 2 + (coordinate1[1] - coordinate2[1]) ** 2

    image_pixels = grid.shape[0]

    image_to_pixelization = np.zeros((image_pixels,))

    for image_index, image_coordinate in enumerate(grid):
        distances = list(map(lambda centers: compute_squared_separation(image_coordinate, centers), pixel_centers))

        image_to_pixelization[image_index] = (np.argmin(distances))

    return image_to_pixelization


@pytest.fixture(name="three_pixels")
def make_three_pixels():
    return np.array([[0, 0], [0, 1], [1, 0]])


@pytest.fixture(name="five_pixels")
def make_five_pixels():
    return np.array([[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]])


class TestMappingMatrix:

    def test__3_image_pixels__6_pixel_pixels__sub_grid_1x1(self, three_pixels):
        sub_to_pixelization = np.array([0, 1, 2])
        sub_to_image = np.array([0, 1, 2])

        grids = MockGridCollection(image=three_pixels, sub=MockSubGridCoords(three_pixels, sub_to_image,
                                                                             sub_grid_size=1))

        pix = mappers.Mapper(pixels=6, grids=grids, pixel_neighbors=None)

        mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pixelization, pix.pixels,
                                                                pix.grids.image.shape[0], pix.grids.sub.sub_to_image,
                                                                pix.grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array([[1, 0, 0, 0, 0, 0],  # Image pixel 0 maps to pix pixel 0.
                                            [0, 1, 0, 0, 0, 0],  # Image pixel 1 maps to pix pixel 1.
                                            [0, 0, 1, 0, 0, 0]])).all()  # Image pixel 2 maps to pix pixel 2

    def test__5_image_pixels__8_pixel_pixels__sub_grid_1x1(self, five_pixels):
        sub_to_pixelization = np.array([0, 1, 2, 7, 6])
        sub_to_image = np.array([0, 1, 2, 3, 4])

        grids = MockGridCollection(image=five_pixels, sub=MockSubGridCoords(five_pixels, sub_to_image,
                                                                            sub_grid_size=1))

        pix = mappers.Mapper(pixels=8, grids=grids, pixel_neighbors=None)

        mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pixelization, pix.pixels,
                                                                pix.grids.image.shape[0], pix.grids.sub.sub_to_image,
                                                                pix.grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[1, 0, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 0 and 3 mappers to pix pixel 0.
             [0, 1, 0, 0, 0, 0, 0, 0],  # Image image_to_pixel 1 and 4 mappers to pix pixel 1.
             [0, 0, 1, 0, 0, 0, 0, 0],
             [0, 0, 0, 0, 0, 0, 0, 1],
             [0, 0, 0, 0, 0, 0, 1, 0]])).all()  # Image image_to_pixel 2 and 5 mappers to pix pixel 2

    def test__5_image_pixels__8_pixel_pixels__sub_grid_2x2__no_overlapping_pixels(self, five_pixels):
        sub_to_pixelization = np.array([0, 1, 2, 3, 1, 2, 3, 4, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
        sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

        grids = MockGridCollection(image=five_pixels, sub=MockSubGridCoords(five_pixels, sub_to_image,
                                                                            sub_grid_size=2))

        pix = mappers.Mapper(pixels=8, grids=grids, pixel_neighbors=None)

        mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pixelization, pix.pixels,
                                                                pix.grids.image.shape[0], pix.grids.sub.sub_to_image,
                                                                pix.grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.25, 0.25, 0.25, 0.25, 0, 0, 0, 0],
             [0, 0.25, 0.25, 0.25, 0.25, 0, 0, 0],
             [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
             [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
             [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

    def test__5_image_pixels__8_pixel_pixels__sub_grid_2x2__include_overlapping_pixels(self, five_pixels):
        sub_to_pixelization = np.array([0, 0, 0, 1, 1, 1, 0, 0, 2, 3, 4, 5, 7, 0, 1, 3, 6, 7, 4, 2])
        sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4])

        grids = MockGridCollection(image=five_pixels, sub=MockSubGridCoords(five_pixels, sub_to_image,
                                                                            sub_grid_size=2))

        pix = mappers.Mapper(pixels=8, grids=grids, pixel_neighbors=None)

        mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pixelization, pix.pixels,
                                                                pix.grids.image.shape[0], pix.grids.sub.sub_to_image,
                                                                pix.grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.75, 0.25, 0, 0, 0, 0, 0, 0],
             [0.5, 0.5, 0, 0, 0, 0, 0, 0],
             [0, 0, 0.25, 0.25, 0.25, 0.25, 0, 0],
             [0.25, 0.25, 0, 0.25, 0, 0, 0, 0.25],
             [0, 0, 0.25, 0, 0.25, 0, 0.25, 0.25]])).all()

    def test__3_image_pixels__6_pixel_pixels__sub_grid_4x4(self, three_pixels):
        sub_to_pixelization = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
                                        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
                                        0, 1, 2, 3, 4, 5, 0, 1, 2, 3, 4, 5, 0, 1, 2, 3])

        sub_to_image = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                                 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])

        grids = MockGridCollection(image=three_pixels, sub=MockSubGridCoords(three_pixels, sub_to_image,
                                                                             sub_grid_size=4))

        pix = mappers.Mapper(pixels=6, grids=grids, pixel_neighbors=None)

        mapping_matrix = pix.mapping_matrix_from_sub_to_pix_jit(sub_to_pixelization, pix.pixels,
                                                                pix.grids.image.shape[0], pix.grids.sub.sub_to_image,
                                                                pix.grids.sub.sub_grid_fraction)

        assert (mapping_matrix == np.array(
            [[0.75, 0.25, 0, 0, 0, 0],
             [0, 0, 1.0, 0, 0, 0],
             [0.1875, 0.1875, 0.1875, 0.1875, 0.125, 0.125]])).all()


class TestRectangularMapper:

    class TestImageAndSubToPixelization:

        def test__3x3_grid_of_pixel_grid__1_coordinate_per_square_pixel__in_centre_of_pixels(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()
            assert pix.pixelization_to_image == [[0], [1], [2], [3], [4], [5], [6], [7], [8]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [3], [4], [5], [6], [7], [8]]

        def test__3x3_grid_of_pixel_grid__1_coordinate_per_square_pixel__near_edges_of_pixels(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [-0.32, -1.0], [-0.32, 0.32], [0.0, 1.0],
                                          [-0.34, -0.34], [-0.34, 0.325], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()
            assert pix.pixelization_to_image == [[0], [1], [2], [3], [4], [5], [6], [7], [8]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [3], [4], [5], [6], [7], [8]]

        def test__3x3_grid_of_pixel_grid__add_multiple_grid_to_1_pixel_pixel(self):
            #                  _ _ _
            # -1.0 to -(1/3)  |_|_|_|
            # -(1/3) to (1/3) |_|_|_|
            #  (1/3) to 1.0   |_|_|_|

            pixelization_grid = np.array([[1.0, -1.0], [0.0, 0.0], [1.0, 1.0],
                                          [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                          [-1.0, -1.0], [0.0, 0.0], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 4, 2, 4, 4, 4, 6, 4, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 4, 2, 4, 4, 4, 6, 4, 8])).all()
            assert pix.pixelization_to_image == [[0], [], [2], [], [1, 3, 4, 5, 7], [], [6], [], [8]]
            assert pix.pixelization_to_sub == [[0], [], [2], [], [1, 3, 4, 5, 7], [], [6], [], [8]]

        def test__4x3_grid_of_pixel_grid__1_coordinate_in_each_pixel(self):
            #   _ _ _
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|

            # Boundaries for column pixel 0 -1.0 to -(1/3)
            # Boundaries for column pixel 1 -(1/3) to (1/3)
            # Boundaries for column pixel 2  (1/3) to 1.0

            # Bounadries for row pixel 0 -1.0 to -0.5
            # Bounadries for row pixel 1 -0.5 to 0.0
            # Bounadries for row pixel 2  0.0 to 0.5
            # Bounadries for row pixel 3  0.5 to 1.0

            pixelization_grid = np.array([[ 1.0, -1.0], [ 1.0,  0.0],
                                          [ 1.0, 1.0], [ 0.5, -1.0],
                                          [-0.5, 1.0], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(4, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=12, shape=(4, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 2, 3, 8, 11])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 3, 8, 11])).all()
            assert pix.pixelization_to_image == [[0], [1], [2], [3], [], [], [], [], [4], [], [], [5]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [3], [], [], [], [], [4], [], [], [5]]

        def test__3x4_grid_of_pixel_grid__1_coordinate_in_each_pixel(self):
            #   _ _ _ _
            #  |_|_|_|_|
            #  |_|_|_|_|
            #  |_|_|_|_|

            # Boundaries for row pixel 0 -1.0 to -(1/3)
            # Boundaries for row pixel 1 -(1/3) to (1/3)
            # Boundaries for row pixel 2  (1/3) to 1.0

            # Bounadries for column pixel 0 -1.0 to -0.5
            # Bounadries for column pixel 1 -0.5 to 0.0
            # Bounadries for column pixel 2  0.0 to 0.5
            # Bounadries for column pixel 3  0.5 to 1.0

            pixelization_grid = np.array([[1.0, -1.0], [1.0, -0.49], [1.0, 0.01], [0.32, 0.01],
                                          [-0.34, -0.01], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 4))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=12, shape=(3, 4), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 2, 6, 9, 11])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 6, 9, 11])).all()
            assert pix.pixelization_to_image == [[0], [1], [2], [], [], [], [3], [], [], [4], [], [5]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [], [], [], [3], [], [], [4], [], [5]]

        def test__3x3_grid__change_arcsecond_dimensions_size__grid_adapts_accordingly(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

            pixelization_grid = np.array([[1.5, -1.5], [1.0, 0.0], [1.0, 0.6], [-1.4, 0.0], [-1.5, 1.5]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 2, 7, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 7, 8])).all()
            assert pix.pixelization_to_image == [[0], [1], [2], [], [], [], [], [3], [4]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [], [], [], [], [3], [4]]

        def test__3x3_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _
            #  |_|_|_| Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
            #  |_|_|_| Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
            #  |_|_|_| Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

            pixelization_grid = np.array([[1.0, -1.5], [1.0, -0.49], [0.32, -1.5], [0.32, 0.51], [-1.0, 1.5]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 3, 5, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 3, 5, 8])).all()
            assert pix.pixelization_to_image == [[0], [1], [], [2], [], [3], [], [], [4]]
            assert pix.pixelization_to_sub == [[0], [1], [], [2], [], [3], [], [], [4]]

        def test__4x3_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|
            #  |_|_|_|

            pixelization_grid = np.array([[1.0, -1.5], [1.0, -0.49], [0.49, -1.5], [-0.6, 0.0], [-1.0, 1.5]])

            pix = pixelizations.Rectangular(shape=(4, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=12, shape=(4, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 3, 10, 11])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 3, 10, 11])).all()
            assert pix.pixelization_to_image == [[0], [1], [], [2], [], [], [], [], [], [], [3], [4]]
            assert pix.pixelization_to_sub == [[0], [1], [], [2], [], [], [], [], [], [], [3], [4]]

        def test__3x4_grid__change_arcsecond_dimensions__not_symmetric(self):
            #   _ _ _ _
            #  |_|_|_|_|
            #  |_|_|_|_|
            #  |_|_|_|_|

            pixelization_grid = np.array([[1.0, -1.5], [1.0, -0.49], [0.32, -1.5], [-0.34, 0.49], [-1.0, 1.5]])

            pix = pixelizations.Rectangular(shape=(3, 4))

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            pix = mappers.RectangularMapper(pixels=12, shape=(3, 4), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 1, 4, 10, 11])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 4, 10, 11])).all()
            assert pix.pixelization_to_image == [[0], [1], [], [], [2], [], [], [], [], [], [3], [4]]
            assert pix.pixelization_to_sub == [[0], [1], [], [], [2], [], [], [], [], [], [3], [4]]

        def test__different_image_and_sub_grids(self):
            #                  _ _ _
            # -1.0 to -(1/3)  |_|_|_|
            # -(1/3) to (1/3) |_|_|_|
            #  (1/3) to 1.0   |_|_|_|

            pixelization_grid = np.array([[1.0, -1.0], [0.0, 0.0], [1.0, 1.0],
                                          [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                          [-1.0, -1.0], [0.0, 0.0], [-1.0, 1.0]])

            pixelization_sub_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                              [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                              [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            grids = MockGridCollection(image=pixelization_grid,
                                       sub=MockSubGridCoords(pixelization_sub_grid, sub_to_image=np.ones((1)),
                                                             sub_grid_size=1))

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=grids, geometry=geometry,
                                            pixel_neighbors=None)

            assert (pix.image_to_pixelization == np.array([0, 4, 2, 4, 4, 4, 6, 4, 8])).all()
            assert (pix.sub_to_pixelization == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()
            assert pix.pixelization_to_image == [[0], [], [2], [], [1, 3, 4, 5, 7], [], [6], [], [8]]
            assert pix.pixelization_to_sub == [[0], [1], [2], [3], [4], [5], [6], [7], [8]]


    class TestReconstructedPixelization:

        def test__3x3_pixelization__solution_vector_ascending(self):

            pixelization_grid = np.array([[2.0, -1.0], [2.0, 0.0], [2.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-2.0, -1.0], [-2.0, 0.0], [-2.0, 1.0]])

            pix = pixelizations.Rectangular(shape=(3, 3))

            geometry = pix.geometry_from_grid(grid=pixelization_grid)

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 3), grids=None, geometry=geometry,
                                            pixel_neighbors=None)

            recon_pix = pix.reconstructed_pixelization_from_solution_vector(solution_vector=np.array([1.0, 2.0, 3.0,
                                                                                                      4.0, 5.0, 6.0,
                                                                                                      7.0, 8.0, 9.0]))

            assert (recon_pix == np.array([[1.0, 2.0, 3.0],
                                           [4.0, 5.0, 6.0],
                                           [7.0, 8.0, 9.0]])).all()
            assert recon_pix.pixel_scales == pytest.approx((4./3., 2./3.), 1e-2)

        def test__compare_to_imaging_util(self):

            pix = mappers.RectangularMapper(pixels=9, shape=(4, 3), grids=None, geometry=MockGeometry(),
                                            pixel_neighbors=None)
            solution = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0])
            recon_pix = pix.reconstructed_pixelization_from_solution_vector(solution_vector=solution)
            recon_pix_util = imaging_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d=solution,
                                                                                                    shape=(4,3))
            assert (recon_pix == recon_pix_util).all()
            assert recon_pix.shape == (4,3)

            pix = mappers.RectangularMapper(pixels=9, shape=(3, 4), grids=None, geometry=MockGeometry(),
                                            pixel_neighbors=None)
            solution = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0])
            recon_pix = pix.reconstructed_pixelization_from_solution_vector(solution_vector=solution)
            recon_pix_util = imaging_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d=solution,
                                                                                                    shape=(3,4))
            assert (recon_pix == recon_pix_util).all()
            assert recon_pix.shape == (3,4)


class TestVoronoiPixMapper:
    class TestImageToPixelizationViaNearestNeighborsForTesting:

        def test__grid_to_pixel_pixels_via_nearest_neighbour__case1__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            image_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1]])

            image_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(image_grid, pixel_centers)

            assert image_to_pixelization[0] == 0
            assert image_to_pixelization[1] == 1
            assert image_to_pixelization[2] == 2
            assert image_to_pixelization[3] == 3

        def test__grid_to_pixel_pixels_via_nearest_neighbour___case2__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            image_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1],
                                   [0.9, -0.9], [-0.9, -0.9], [-0.9, 0.9], [0.9, 0.9]])

            image_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(image_grid, pixel_centers)

            assert image_to_pixelization[0] == 0
            assert image_to_pixelization[1] == 1
            assert image_to_pixelization[2] == 2
            assert image_to_pixelization[3] == 3
            assert image_to_pixelization[4] == 3
            assert image_to_pixelization[5] == 2
            assert image_to_pixelization[6] == 1
            assert image_to_pixelization[7] == 0

        def test__grid_to_pixel_pixels_via_nearest_neighbour___case3__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0], [0.0, 0.0], [2.0, 2.0]])
            image_grid = np.array([[0.1, 0.1], [-0.1, -0.1], [0.49, 0.49],
                                   [0.51, 0.51], [1.01, 1.01], [1.51, 1.51]])

            image_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(image_grid, pixel_centers)

            assert image_to_pixelization[0] == 4
            assert image_to_pixelization[1] == 4
            assert image_to_pixelization[2] == 4
            assert image_to_pixelization[3] == 0
            assert image_to_pixelization[4] == 0
            assert image_to_pixelization[5] == 5

    class TestSubToPixelizationViaNearestNeighborsForTesting:

        def test__grid_to_pixel_pixels_via_nearest_neighbour__case1__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            sub_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1]])

            sub_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(sub_grid, pixel_centers)

            assert sub_to_pixelization[0] == 0
            assert sub_to_pixelization[1] == 1
            assert sub_to_pixelization[2] == 2
            assert sub_to_pixelization[3] == 3

        def test__grid_to_pixel_pixels_via_nearest_neighbour___case2__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
            sub_grid = np.array([[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1],
                                 [0.9, -0.9], [-0.9, -0.9], [-0.9, 0.9], [0.9, 0.9]])

            sub_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(sub_grid, pixel_centers)

            assert sub_to_pixelization[0] == 0
            assert sub_to_pixelization[1] == 1
            assert sub_to_pixelization[2] == 2
            assert sub_to_pixelization[3] == 3
            assert sub_to_pixelization[4] == 3
            assert sub_to_pixelization[5] == 2
            assert sub_to_pixelization[6] == 1
            assert sub_to_pixelization[7] == 0

        def test__grid_to_pixel_pixels_via_nearest_neighbour___case3__correct_pairs(self):
            pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0], [0.0, 0.0], [2.0, 2.0]])
            sub_grid = np.array([[0.1, 0.1], [-0.1, -0.1], [0.49, 0.49],
                                 [0.51, 0.51], [1.01, 1.01], [1.51, 1.51]])

            sub_to_pixelization = grid_to_pixel_pixels_via_nearest_neighbour(sub_grid, pixel_centers)

            assert sub_to_pixelization[0] == 4
            assert sub_to_pixelization[1] == 4
            assert sub_to_pixelization[2] == 4
            assert sub_to_pixelization[3] == 0
            assert sub_to_pixelization[4] == 0
            assert sub_to_pixelization[5] == 5

    class TestImageToPixelization:

        def test__image_to_pixelization_of_mapper_matches_nearest_neighbor_calculation(self):
            pixel_centers = np.array([[-1.0, -1.0], [-0.9, 0.9],
                                      [1.0, -1.1], [1.2, 1.2]])

            pixelization_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            image_to_pixelization_nearest_neighbour = grid_to_pixel_pixels_via_nearest_neighbour(pixelization_grid,
                                                                                                 pixel_centers)

            sub_to_image = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
            grids = MockGridCollection(image=pixelization_grid, sub=MockSubGridCoords(np.array([]), sub_to_image,
                                                                                      sub_grid_size=1))

            pix = pixelizations.Voronoi(pixels=6)
            voronoi = pix.voronoi_from_pixel_centers(pixel_centers)
            pixel_neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)
            voronoi_to_pixelization = np.array([0, 1, 2, 3])
            image_to_voronoi = np.array([0, 0, 1, 0, 0, 1, 2, 2, 3])

            mapper = mappers.VoronoiMapper(pixels=6, grids=grids, pixel_neighbors=pixel_neighbors,
                                           pixel_centers=pixel_centers, voronoi=voronoi,
                                           voronoi_to_pixelization=voronoi_to_pixelization,
                                           image_to_voronoi=image_to_voronoi)

            assert (mapper.image_to_pixelization == image_to_pixelization_nearest_neighbour).all()

    class TestSubToPixelization:

        def test__sub_to_pixelization_of_mapper_matches_nearest_neighbor_calculation(self):
            pixel_centers = np.array([[0.1, 0.1], [1.1, 0.1], [2.1, 0.1],
                                      [0.1, 1.1], [1.1, 1.1], [2.1, 1.1]])

            pixelization_sub_grid = np.array([[0.05, 0.15], [0.15, 0.15], [0.05, 0.05], [0.15, 0.05],
                                              [1.05, 0.15], [1.15, 0.15], [1.05, 0.05], [1.15, 0.05],
                                              [2.05, 0.15], [2.15, 0.15], [2.05, 0.05], [2.15, 0.05],
                                              [0.05, 1.15], [0.15, 1.15], [0.05, 1.05], [0.15, 1.05],
                                              [1.05, 1.15], [1.15, 1.15], [1.05, 1.05], [1.15, 1.05],
                                              [2.05, 1.15], [2.15, 1.15], [2.05, 1.05], [2.15, 1.05]])

            sub_to_pixelization_nearest_neighbour = grid_to_pixel_pixels_via_nearest_neighbour(pixelization_sub_grid,
                                                                                               pixel_centers)

            sub_to_image = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5])
            grids = MockGridCollection(image=np.array([]), sub=MockSubGridCoords(pixelization_sub_grid, sub_to_image,
                                                                                 sub_grid_size=1))

            pix = pixelizations.Voronoi(pixels=6)
            voronoi = pix.voronoi_from_pixel_centers(pixel_centers)
            pixel_neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)
            voronoi_to_pixelization = np.array([0, 1, 2, 3])
            image_to_voronoi = np.array([0, 0, 1, 0, 0, 1, 2, 2, 3])

            mapper = mappers.VoronoiMapper(pixels=6, grids=grids, pixel_neighbors=pixel_neighbors,
                                           pixel_centers=pixel_centers, voronoi=voronoi,
                                           voronoi_to_pixelization=voronoi_to_pixelization,
                                           image_to_voronoi=image_to_voronoi)

            assert (mapper.sub_to_pixelization == sub_to_pixelization_nearest_neighbour).all()



import numpy as np

from autolens.inversion import regularization


class TestRegularizationConstant:
    class TestRegularizationMatrixFromNeighbors:

        def test__1_b_matrix_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
            # Here, we define the pixel_neighbors first here and make the B matrices based on them.

            # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
            # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
            # the rows of each B matrix wherever we like ;0.

            pixel_neighbors = np.array([[1, 2], [0], [0]])

            test_b_matrix = np.array([[-1, 1, 0],  # Pair 1
                                      [-1, 0, 1],  # Pair 2
                                      [0, 0, 0, ]])  # Pair 1 flip

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(3)

            reg = regularization.Constant(coefficients=(1.0,))
            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            test_b_matrix = np.array([[-1, 1, 0, 0],
                                      [0, -1, 1, 0],
                                      [0, 0, -1, 1],
                                      [1, 0, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

            pixel_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])

            reg = regularization.Constant(coefficients=(1.0,))
            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_4x4__coefficient_2__makes_correct_regularization_matrix(self):
            pixel_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])

            test_b_matrix = 2.0 * np.array([[-1, 1, 0, 0],
                                            [0, -1, 1, 0],
                                            [0, 0, -1, 1],
                                            [1, 0, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

            reg = regularization.Constant(coefficients=(2.0,))
            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_9x9__coefficient_2__makes_correct_regularization_matrix(self):
            pixel_neighbors = [[1, 3], [4, 2, 0], [1, 5], [4, 6, 0], [7, 1, 5, 3], [4, 2, 8], [7, 3], [4, 8, 6],
                               [7, 5]]

            test_b_matrix_0 = np.array([[-1, 1, 0, 0, 0, 0, 0, 0, 0],
                                        [-1, 0, 0, 1, 0, 0, 0, 0, 0],
                                        [0, -1, 1, 0, 0, 0, 0, 0, 0],
                                        [0, -1, 0, 0, 1, 0, 0, 0, 0],
                                        [0, 0, -1, 0, 0, 1, 0, 0, 0],
                                        [0, 0, 0, -1, 1, 0, 0, 0, 0],
                                        [0, 0, 0, -1, 0, 0, 1, 0, 0],
                                        [0, 0, 0, 0, -1, 1, 0, 0, 0],
                                        [0, 0, 0, 0, -1, 0, 0, 1, 0]])

            test_b_matrix_1 = np.array([[0, 0, 0, 0, 0, -1, 0, 0, 1],
                                        [0, 0, 0, 0, 0, 0, -1, 1, 0],
                                        [0, 0, 0, 0, 0, 0, 0, -1, 1],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_0 = np.matmul(test_b_matrix_0.T, test_b_matrix_0)
            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_regularization_matrix = test_regularization_matrix_0 + \
                                         test_regularization_matrix_1 + 1e-8 * np.identity(9)

            reg = regularization.Constant(coefficients=(1.0,))
            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)


class TestRegularizationWeighted:
    class TestPixelSignals:

        def test__x3_image_pixels_signals_1s__pixel_scale_1__pixel_signals_all_1s(self):
            reg = regularization.Weighted(signal_scale=1.0)

            image_to_pixelization = np.array([0, 1, 2])
            galaxy_image = np.array([1.0, 1.0, 1.0])

            pixel_signals = reg.pixel_signals_from_images(pixels=3, image_to_pixelization=image_to_pixelization,
                                                          galaxy_image=galaxy_image)

            assert (pixel_signals == np.array([1.0, 1.0, 1.0])).all()

        def test__x4_image_pixels_signals_1s__pixel_signals_still_all_1s(self):
            reg = regularization.Weighted(signal_scale=1.0)

            image_to_pixelization = np.array([0, 1, 2, 0])
            galaxy_image = np.array([1.0, 1.0, 1.0, 1.0])

            pixel_signals = reg.pixel_signals_from_images(pixels=3, image_to_pixelization=image_to_pixelization,
                                                          galaxy_image=galaxy_image)

            assert (pixel_signals == np.array([1.0, 1.0, 1.0])).all()

        def test__galaxy_flux_in_a_pixel_pixel_is_double_the_others__pixel_signal_is_1_others_a_half(self):
            reg = regularization.Weighted(signal_scale=1.0)

            image_to_pixelization = np.array([0, 1, 2])
            galaxy_image = np.array([2.0, 1.0, 1.0])

            pixel_signals = reg.pixel_signals_from_images(pixels=3, image_to_pixelization=image_to_pixelization,
                                                          galaxy_image=galaxy_image)

            assert (pixel_signals == np.array([1.0, 0.5, 0.5])).all()

        def test__same_as_above_but_pixel_scale_2__scales_pixel_signals(self):
            reg = regularization.Weighted(signal_scale=2.0)

            image_to_pixelization = np.array([0, 1, 2])
            galaxy_image = np.array([2.0, 1.0, 1.0])

            pixel_signals = reg.pixel_signals_from_images(pixels=3, image_to_pixelization=image_to_pixelization,
                                                          galaxy_image=galaxy_image)

            assert (pixel_signals == np.array([1.0, 0.25, 0.25])).all()

    class TestRegularizationWeights(object):

        def test__pixel_signals_all_1s__coefficients_all_1s__weights_all_1s(self):
            reg = regularization.Weighted(coefficients=(1.0, 1.0))

            pixel_signals = np.array([1.0, 1.0, 1.0])

            weights = reg.regularization_weights_from_pixel_signals(pixel_signals)

            assert (weights == np.array([1.0, 1.0, 1.0])).all()

        def test__pixel_signals_vary__coefficents_all_1s__weights_still_all_1s(self):
            reg = regularization.Weighted(coefficients=(1.0, 1.0))

            pixel_signals = np.array([0.25, 0.5, 0.75])

            weights = reg.regularization_weights_from_pixel_signals(pixel_signals)

            assert (weights == np.array([1.0, 1.0, 1.0])).all()

        def test__pixel_signals_vary__coefficents_1_and_0__weights_are_pixel_signals_squared(self):
            reg = regularization.Weighted(coefficients=(1.0, 0.0))

            pixel_signals = np.array([0.25, 0.5, 0.75])

            weights = reg.regularization_weights_from_pixel_signals(pixel_signals)

            assert (weights == np.array([0.25 ** 2.0, 0.5 ** 2.0, 0.75 ** 2.0])).all()

        def test__pixel_signals_vary__coefficents_0_and_1__weights_are_1_minus_pixel_signals_squared(self):
            reg = regularization.Weighted(coefficients=(0.0, 1.0))

            pixel_signals = np.array([0.25, 0.5, 0.75])

            weights = reg.regularization_weights_from_pixel_signals(pixel_signals)

            assert (weights == np.array([0.75 ** 2.0, 0.5 ** 2.0, 0.25 ** 2.0])).all()

    class TestRegularizationMatrix:

        def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            reg = regularization.Weighted()

            pixel_neighbors = np.array([[2], [3], [0], [1]])

            test_b_matrix = np.array([[-1, 0, 1, 0],
                                      [0, -1, 0, 1],
                                      [1, 0, -1, 0],
                                      [0, 1, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix)

            regularization_weights = np.ones((4,))

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
            # Here, we define the pixel_neighbors first here and make the B matrices based on them.

            # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
            # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
            # the rows of each B matrix wherever we like ;0.

            reg = regularization.Weighted()

            pixel_neighbors = np.array([[1, 2], [0], [0]])

            test_b_matrix_1 = np.array([[-1, 1, 0],  # Pair 1
                                        [-1, 0, 1],  # Pair 2
                                        [1, -1, 0]])  # Pair 1 flip

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[1, 0, -1],  # Pair 2 flip
                                        [0, 0, 0],
                                        [0, 0, 0]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            regularization_weights = np.ones((3))

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            reg = regularization.Weighted()

            test_b_matrix_1 = np.array([[-1, 1, 0, 0],
                                        [0, -1, 1, 0],
                                        [0, 0, -1, 1],
                                        [1, 0, 0, -1]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[-1, 0, 0, 1],
                                        [1, -1, 0, 0],
                                        [0, 1, -1, 0],
                                        [0, 0, 1, -1]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            pixel_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])
            regularization_weights = np.ones((4,))

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__4_b_matrices_size_6x6__weights_all_1s__makes_correct_regularization_matrix(self):
            # Again, lets exploit the freedom we have when setting up our B matrices to make matching it to pairs a
            # lot less Stressful.

            reg = regularization.Weighted()

            pixel_neighbors = [[2, 3, 4], [2, 5], [0, 1, 3, 5], [0, 2], [5, 0], [4, 1, 2]]

            test_b_matrix_1 = np.array([[-1, 0, 1, 0, 0, 0],  # Pair 1
                                        [0, -1, 1, 0, 0, 0],  # Pair 2
                                        [-1, 0, 0, 1, 0, 0],  # Pair 3
                                        [0, 0, 0, 0, -1, 1],  # Pair 4
                                        [0, -1, 0, 0, 0, 1],  # Pair 5
                                        [-1, 0, 0, 0, 1, 0]])  # Pair 6

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[0, 0, -1, 1, 0, 0],  # Pair 7
                                        [0, 0, -1, 0, 0, 1],  # Pair 8
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_b_matrix_3 = np.array([[1, 0, -1, 0, 0, 0],  # Pair 1 flip
                                        [0, 1, -1, 0, 0, 0],  # Pair 2 flip
                                        [1, 0, 0, -1, 0, 0],  # Pair 3 flip
                                        [0, 0, 0, 0, 1, -1],  # Pair 4 flip
                                        [0, 1, 0, 0, 0, -1],  # Pair 5 flip
                                        [1, 0, 0, 0, -1, 0]])  # Pair 6 flip

            test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)

            test_b_matrix_4 = np.array([[0, 0, 1, -1, 0, 0],  # Pair 7 flip
                                        [0, 0, 1, 0, 0, -1],  # Pair 8 flip
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + \
                                         test_regularization_matrix_3 + + test_regularization_matrix_4

            regularization_weights = np.ones((6))

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__1_b_matrix_size_3x3_variables_regularization_weights__makes_correct_regularization_matrix(self):
            # Simple case, where we have just one regularization direction, regularizing pixel 0 -> 1 and 1 -> 2.

            # This means our B matrix is:

            # [-1, 1, 0]
            # [0, -1, 1]
            # [0, 0, -1]

            # Regularization Matrix, H = B * B.T.I can

            reg = regularization.Weighted()

            regularization_weights = np.array([2.0, 4.0, 1.0])

            test_b_matrix = np.array([[-1, 1, 0],  # [[-2, 2, 0], (Matrix)
                                      [1, -1, 0],  # [4, -4, 0], (after)
                                      [0, 0, 0]])  # [0, 0,  0]]) (weights)

            test_b_matrix = (test_b_matrix.T * regularization_weights).T

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix)

            pixel_neighbors = [[1], [0], []]

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_4x4_variables_regularization_weights__makes_correct_regularization_matrix(self):
            # Simple case, where we have just one regularization direction, regularizing pixel 0 -> 1 and 1 -> 2.

            # This means our B matrix is:

            # [-1, 1, 0]
            # [0, -1, 1]
            # [0, 0, -1]

            # Regularization Matrix, H = B * B.T.I can

            reg = regularization.Weighted()

            regularization_weights = np.array([2.0, 4.0, 1.0, 8.0])

            test_b_matrix_1 = np.array([[-2, 2, 0, 0],
                                        [-2, 0, 2, 0],
                                        [0, -4, 4, 0],
                                        [0, -4, 0, 4]])

            test_b_matrix_2 = np.array([[4, -4, 0, 0],
                                        [1, 0, -1, 0],
                                        [0, 1, -1, 0],
                                        [0, 8, 0, -8]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            pixel_neighbors = [[1, 2], [0, 2, 3], [0, 1], [1]]

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__4_b_matrices_size_6x6_with_regularization_weights__makes_correct_regularization_matrix(self):
            reg = regularization.Weighted()

            pixel_neighbors = [[1, 4], [2, 4, 0], [3, 4, 5, 1], [5, 2], [5, 0, 1, 2], [2, 3, 4]]
            regularization_weights = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])

            # I'm inputting the regularization weights directly thiss time, as it'd be a pain to multiply with a
            # loop.

            test_b_matrix_1 = np.array([[-1, 1, 0, 0, 0, 0],  # Pair 1
                                        [-1, 0, 0, 0, 1, 0],  # Pair 2
                                        [0, -2, 2, 0, 0, 0],  # Pair 3
                                        [0, -2, 0, 0, 2, 0],  # Pair 4
                                        [0, 0, -3, 3, 0, 0],  # Pair 5
                                        [0, 0, -3, 0, 3, 0]])  # Pair 6

            test_b_matrix_2 = np.array([[0, 0, -3, 0, 0, 3],  # Pair 7
                                        [0, 0, 0, -4, 0, 4],  # Pair 8
                                        [0, 0, 0, 0, -5, 5],  # Pair 9
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            # Now do the same pairs but with the regularization direction and weights swapped.

            test_b_matrix_3 = np.array([[2, -2, 0, 0, 0, 0],  # Pair 1
                                        [5, 0, 0, 0, -5, 0],  # Pair 2
                                        [0, 3, -3, 0, 0, 0],  # Pair 3
                                        [0, 5, 0, 0, -5, 0],  # Pair 4
                                        [0, 0, 4, -4, 0, 0],  # Pair 5
                                        [0, 0, 5, 0, -5, 0]])  # Pair 6

            test_b_matrix_4 = np.array([[0, 0, 6, 0, 0, -6],  # Pair 7
                                        [0, 0, 0, 6, 0, -6],  # Pair 8
                                        [0, 0, 0, 0, 6, -6],  # Pair 9
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)
            test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)
            test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + \
                                         test_regularization_matrix_3 + test_regularization_matrix_4

            regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(regularization_weights,
                                                                                   pixel_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
